@@ -47,9 +47,9 @@ defmodule Registry do
 
       {:ok, _} = Registry.start_link(keys: :unique, name: Registry.ViaTest)
       name = {:via, Registry, {Registry.ViaTest, "agent", :hello}}
-      {:ok, _} = Agent.start_link(fn -> 0 end, name: name)
+      {:ok, agent_pid} = Agent.start_link(fn -> 0 end, name: name)
       Registry.lookup(Registry.ViaTest, "agent")
-      #=> [{self(), :hello}]
+      #=> [{agent_pid, :hello}]
 
   To this point, we have been starting `Registry` using `start_link/1`.
   Typically the registry is started as part of a supervision tree though:
@@ -165,7 +165,7 @@ defmodule Registry do
   the value from the registry and sending it a message. Many parts of the standard
   library are designed to cope with that, such as `Process.monitor/1` which will
   deliver the `:DOWN` message immediately if the monitored process is already dead
-  and `Kernel.send/2` which acts as a no-op for dead processes.
+  and `send/2` which acts as a no-op for dead processes.
 
   ## ETS
 
@@ -217,6 +217,16 @@ defmodule Registry do
           | {:listeners, [atom]}
           | {:meta, [{meta_key, meta_value}]}
 
+  @typedoc """
+  The message that the registry sends to listeners when a process registers or unregisters.
+
+  See the `:listeners` option in `start_link/1`.
+  """
+  @typedoc since: "1.15.0"
+  @type listener_message ::
+          {:register, registry, key, registry_partition :: pid, value}
+          | {:unregister, registry, key, registry_partition :: pid}
+
   ## Via callbacks
 
   @doc false
@@ -260,6 +270,10 @@ defmodule Registry do
     end
   end
 
+  def send({registry, key, _value}, msg) do
+    Registry.send({registry, key}, msg)
+  end
+
   @doc false
   def unregister_name({registry, key}), do: unregister(registry, key)
   def unregister_name({registry, key, _value}), do: unregister(registry, key)
@@ -305,10 +319,10 @@ defmodule Registry do
   The following keys are optional:
 
     * `:partitions` - the number of partitions in the registry. Defaults to `1`.
-    * `:listeners` - a list of named processes which are notified of `:register`
-      and `:unregister` events. The registered process must be monitored by the
+    * `:listeners` - a list of named processes which are notified of register
+      and unregister events. The registered process must be monitored by the
       listener if the listener wants to be notified if the registered process
-      crashes.
+      crashes. Messages sent to listeners are of type `t:listener_message/0`.
     * `:meta` - a keyword list of metadata to be attached to the registry.
 
   """
@@ -354,13 +368,20 @@ defmodule Registry do
             "expected :listeners to be a list of named processes, got: #{inspect(listeners)}"
     end
 
+    compressed = Keyword.get(options, :compressed, false)
+
+    unless is_boolean(compressed) do
+      raise ArgumentError,
+            "expected :compressed to be a boolean, got: #{inspect(compressed)}"
+    end
+
     # The @info format must be kept in sync with Registry.Partition optimization.
     entries = [
       {@all_info, {keys, partitions, nil, nil, listeners}},
       {@key_info, {keys, partitions, nil}} | meta
     ]
 
-    Registry.Supervisor.start_link(keys, name, partitions, listeners, entries)
+    Registry.Supervisor.start_link(keys, name, partitions, listeners, entries, compressed)
   end
 
   @doc false
@@ -724,6 +745,64 @@ defmodule Registry do
     acc
   end
 
+  @doc """
+  Reads the values for the given `key` for `pid` in `registry`.
+
+  For unique registries, it is either an empty list or a list
+  with a single element. For duplicate registries, it is a list
+  with zero, one, or multiple elements.
+
+  ## Examples
+
+  In the example below we register the current process and look it up
+  both from itself and other processes:
+
+      iex> Registry.start_link(keys: :unique, name: Registry.UniqueLookupTest)
+      iex> Registry.values(Registry.UniqueLookupTest, "hello", self())
+      []
+      iex> {:ok, _} = Registry.register(Registry.UniqueLookupTest, "hello", :world)
+      iex> Registry.values(Registry.UniqueLookupTest, "hello", self())
+      [:world]
+      iex> Task.async(fn -> Registry.values(Registry.UniqueLookupTest, "hello", self()) end) |> Task.await()
+      []
+      iex> parent = self()
+      iex> Task.async(fn -> Registry.values(Registry.UniqueLookupTest, "hello", parent) end) |> Task.await()
+      [:world]
+
+  The same applies to duplicate registries:
+
+      iex> Registry.start_link(keys: :duplicate, name: Registry.DuplicateLookupTest)
+      iex> Registry.values(Registry.DuplicateLookupTest, "hello", self())
+      []
+      iex> {:ok, _} = Registry.register(Registry.DuplicateLookupTest, "hello", :world)
+      iex> Registry.values(Registry.DuplicateLookupTest, "hello", self())
+      [:world]
+      iex> {:ok, _} = Registry.register(Registry.DuplicateLookupTest, "hello", :another)
+      iex> Enum.sort(Registry.values(Registry.DuplicateLookupTest, "hello", self()))
+      [:another, :world]
+
+  """
+  @doc since: "1.12.0"
+  @spec values(registry, key, pid) :: [value]
+  def values(registry, key, pid) when is_atom(registry) do
+    case key_info!(registry) do
+      {:unique, partitions, key_ets} ->
+        key_ets = key_ets || key_ets!(registry, key, partitions)
+
+        case safe_lookup_second(key_ets, key) do
+          {^pid, value} ->
+            [value]
+
+          _ ->
+            []
+        end
+
+      {:duplicate, partitions, key_ets} ->
+        key_ets = key_ets || key_ets!(registry, pid, partitions)
+        for {^pid, value} <- safe_lookup_second(key_ets, key), do: value
+    end
+  end
+
   @doc """
   Unregisters all entries for the given `key` associated to the current
   process in `registry`.
@@ -732,6 +811,10 @@ defmodule Registry do
   the owner if there are no more keys associated to the current process. See
   also `register/3` to read more about the "owner".
 
+  If the registry has listeners specified via the `:listeners` option in `start_link/1`,
+  those listeners will be notified of the unregistration and will receive a
+  message of type `t:listener_message/0`.
+
   ## Examples
 
   For unique registries:
@@ -893,6 +976,10 @@ defmodule Registry do
   If the registry has duplicate keys, multiple registrations from the
   current process under the same key are allowed.
 
+  If the registry has listeners specified via the `:listeners` option in `start_link/1`,
+  those listeners will be notified of the registration and will receive a
+  message of type `t:listener_message/0`.
+
   ## Examples
 
   Registering under a unique registry does not allow multiple entries:
@@ -996,7 +1083,9 @@ defmodule Registry do
       :ets.lookup(registry, key)
     catch
       :error, :badarg ->
-        raise ArgumentError, "unknown registry: #{inspect(registry)}"
+        raise ArgumentError,
+              "unknown registry: #{inspect(registry)}. Either the registry name is invalid or " <>
+                "the registry is not running, possibly because its application isn't started"
     else
       [{^key, value}] -> {:ok, value}
       _ -> :error
@@ -1207,7 +1296,7 @@ defmodule Registry do
 
   ## Examples
 
-  This example shows how to get everything from the registry.
+  This example shows how to get everything from the registry:
 
       iex> Registry.start_link(keys: :unique, name: Registry.SelectAllTest)
       iex> {:ok, _} = Registry.register(Registry.SelectAllTest, "hello", :value)
@@ -1215,7 +1304,7 @@ defmodule Registry do
       iex> Registry.select(Registry.SelectAllTest, [{{:"$1", :"$2", :"$3"}, [], [{{:"$1", :"$2", :"$3"}}]}])
       [{"world", self(), :value}, {"hello", self(), :value}]
 
-  Get all keys in the registry.
+  Get all keys in the registry:
 
       iex> Registry.start_link(keys: :unique, name: Registry.SelectAllTest)
       iex> {:ok, _} = Registry.register(Registry.SelectAllTest, "hello", :value)
@@ -1228,17 +1317,7 @@ defmodule Registry do
   @spec select(registry, spec) :: [term]
   def select(registry, spec)
       when is_atom(registry) and is_list(spec) do
-    spec =
-      for part <- spec do
-        case part do
-          {{key, pid, value}, guards, select} ->
-            {{key, {pid, value}}, guards, select}
-
-          _ ->
-            raise ArgumentError,
-                  "invalid match specification in Registry.select/2: #{inspect(spec)}"
-        end
-      end
+    spec = group_match_headers(spec, __ENV__.function)
 
     case key_info!(registry) do
       {_kind, partitions, nil} ->
@@ -1251,6 +1330,51 @@ defmodule Registry do
     end
   end
 
+  @doc """
+  Works like `select/2`, but only returns the number of matching records.
+
+  ## Examples
+
+  In the example below we register the current process under different
+  keys in a unique registry but with the same value:
+
+      iex> Registry.start_link(keys: :unique, name: Registry.CountSelectTest)
+      iex> {:ok, _} = Registry.register(Registry.CountSelectTest, "hello", :value)
+      iex> {:ok, _} = Registry.register(Registry.CountSelectTest, "world", :value)
+      iex> Registry.count_select(Registry.CountSelectTest, [{{:_, :_, :value}, [], [true]}])
+      2
+  """
+  @doc since: "1.14.0"
+  @spec count_select(registry, spec) :: non_neg_integer()
+  def count_select(registry, spec)
+      when is_atom(registry) and is_list(spec) do
+    spec = group_match_headers(spec, __ENV__.function)
+
+    case key_info!(registry) do
+      {_kind, partitions, nil} ->
+        Enum.reduce(0..(partitions - 1), 0, fn partition_index, acc ->
+          count = :ets.select_count(key_ets!(registry, partition_index), spec)
+          acc + count
+        end)
+
+      {_kind, 1, key_ets} ->
+        :ets.select_count(key_ets, spec)
+    end
+  end
+
+  defp group_match_headers(spec, {fun, arity}) do
+    for part <- spec do
+      case part do
+        {{key, pid, value}, guards, select} ->
+          {{key, {pid, value}}, guards, select}
+
+        _ ->
+          raise ArgumentError,
+                "invalid match specification in Registry.#{fun}/#{arity}: #{inspect(spec)}"
+      end
+    end
+  end
+
   ## Helpers
 
   @compile {:inline, hash: 2}
@@ -1339,12 +1463,12 @@ defmodule Registry.Supervisor do
   @moduledoc false
   use Supervisor
 
-  def start_link(kind, registry, partitions, listeners, entries) do
-    arg = {kind, registry, partitions, listeners, entries}
+  def start_link(kind, registry, partitions, listeners, entries, compressed) do
+    arg = {kind, registry, partitions, listeners, entries, compressed}
     Supervisor.start_link(__MODULE__, arg, name: registry)
   end
 
-  def init({kind, registry, partitions, listeners, entries}) do
+  def init({kind, registry, partitions, listeners, entries, compressed}) do
     ^registry = :ets.new(registry, [:set, :public, :named_table, read_concurrency: true])
     true = :ets.insert(registry, entries)
 
@@ -1352,7 +1476,7 @@ defmodule Registry.Supervisor do
       for i <- 0..(partitions - 1) do
         key_partition = Registry.Partition.key_name(registry, i)
         pid_partition = Registry.Partition.pid_name(registry, i)
-        arg = {kind, registry, i, partitions, key_partition, pid_partition, listeners}
+        arg = {kind, registry, i, partitions, key_partition, pid_partition, listeners, compressed}
 
         %{
           id: pid_partition,
@@ -1412,9 +1536,9 @@ defmodule Registry.Partition do
 
   ## Callbacks
 
-  def init({kind, registry, i, partitions, key_partition, pid_partition, listeners}) do
+  def init({kind, registry, i, partitions, key_partition, pid_partition, listeners, compressed}) do
     Process.flag(:trap_exit, true)
-    key_ets = init_key_ets(kind, key_partition)
+    key_ets = init_key_ets(kind, key_partition, compressed)
     pid_ets = init_pid_ets(kind, pid_partition)
 
     # If we have only one partition, we do an optimization which
@@ -1435,17 +1559,18 @@ defmodule Registry.Partition do
 
   # The key partition is a set for unique keys,
   # duplicate bag for duplicate ones.
-  defp init_key_ets(:unique, key_partition) do
-    :ets.new(key_partition, [:set, :public, read_concurrency: true, write_concurrency: true])
+  defp init_key_ets(:unique, key_partition, compressed) do
+    opts = [:set, :public, read_concurrency: true, write_concurrency: true]
+    :ets.new(key_partition, compression_opt(opts, compressed))
   end
 
-  defp init_key_ets(:duplicate, key_partition) do
-    :ets.new(key_partition, [
-      :duplicate_bag,
-      :public,
-      read_concurrency: true,
-      write_concurrency: true
-    ])
+  defp init_key_ets(:duplicate, key_partition, compressed) do
+    opts = [:duplicate_bag, :public, read_concurrency: true, write_concurrency: true]
+    :ets.new(key_partition, compression_opt(opts, compressed))
+  end
+
+  defp compression_opt(opts, compressed) do
+    if compressed, do: [:compressed] ++ opts, else: opts
   end
 
   # A process can always have multiple keys, so the
