@@ -246,7 +246,7 @@ defmodule Module.Types.IntegrationTest do
       assert_warnings(files, warning)
     end
 
-    test "doesn't report missing funcs at compile time" do
+    test "doesn't report missing functions at compile time" do
       files = %{
         "a.ex" => """
         Enum.map([], fn _ -> BadReferencer.no_func4() end)
@@ -576,6 +576,22 @@ defmodule Module.Types.IntegrationTest do
                |> Code.compile_quoted()
              end) =~ "warning:"
     end
+
+    test "do not parse binary segments as variables" do
+      files = %{
+        "a.ex" => """
+        defmodule A do
+          def decode(byte) do
+            case byte do
+              enc when enc in [<<0x00>>, <<0x01>>] -> :ok
+            end
+          end
+        end
+        """
+      }
+
+      assert_no_warnings(files)
+    end
   end
 
   describe "after_verify" do
@@ -701,7 +717,7 @@ defmodule Module.Types.IntegrationTest do
 
       warning = """
       warning: A.a/0 is deprecated. oops
-        b.ex:3: B
+        b.ex:3: B (module)
 
       """
 
@@ -783,7 +799,7 @@ defmodule Module.Types.IntegrationTest do
   end
 
   defp read_chunk(binary) do
-    assert {:ok, {_module, [{'ExCk', chunk}]}} = :beam_lib.chunks(binary, ['ExCk'])
+    assert {:ok, {_module, [{~c"ExCk", chunk}]}} = :beam_lib.chunks(binary, [~c"ExCk"])
     assert {:elixir_checker_v1, map} = :erlang.binary_to_term(chunk)
     map
   end
