@@ -1,12 +1,12 @@
 defmodule Mix.Compilers.Elixir do
   @moduledoc false
 
-  @manifest_vsn 14
+  @manifest_vsn 20
   @checkpoint_vsn 2
 
   import Record
 
-  defrecord :module, [:module, :kind, :sources, :export, :recompile?]
+  defrecord :module, [:module, :kind, :sources, :export, :recompile?, :timestamp]
 
   defrecord :source,
     source: nil,
@@ -17,7 +17,8 @@ defmodule Mix.Compilers.Elixir do
     runtime_references: [],
     compile_env: [],
     external: [],
-    warnings: [],
+    compile_warnings: [],
+    runtime_warnings: [],
     modules: []
 
   @doc """
@@ -32,6 +33,7 @@ defmodule Mix.Compilers.Elixir do
   have changed at runtime.
   """
   def compile(manifest, srcs, dest, new_cache_key, new_parent_manifests, new_parents, opts) do
+    Mix.ensure_application!(:crypto)
     modified = Mix.Utils.last_modified(manifest)
     new_parents = :ordsets.from_list(new_parents)
 
@@ -44,6 +46,11 @@ defmodule Mix.Compilers.Elixir do
     {all_modules, all_sources, all_local_exports, old_parents, old_cache_key, old_deps_config} =
       parse_manifest(manifest, dest)
 
+    # Prepend ourselves early because of __mix_recompile__? checks
+    # and also that, in case of nothing compiled, we already need
+    # ourselves available in the path.
+    Code.prepend_path(dest)
+
     # If modules have been added or removed from the Erlang compiler,
     # we need to recompile all references to old and new modules.
     stale =
@@ -54,43 +61,48 @@ defmodule Mix.Compilers.Elixir do
         []
       end
 
+    local_deps = Enum.reject(Mix.Dep.cached(), & &1.scm.fetchable?())
+
     # If mix.exs has changed, recompile anything that calls Mix.Project.
     stale =
       if Mix.Utils.stale?([Mix.Project.project_file()], [modified]),
         do: [Mix.Project | stale],
         else: stale
 
-    # If the dependencies have changed, we need to traverse lock/config files.
-    deps_changed? = Mix.Utils.stale?([Mix.Project.config_mtime()], [modified])
+    # If the lock has changed or a local dependency was added ore removed,
+    # we need to traverse lock/config files.
+    deps_changed? =
+      Mix.Utils.stale?([Mix.Project.config_mtime()], [modified]) or
+        local_deps_changed?(old_deps_config, local_deps)
 
     # If a configuration is only accessed at compile-time, we don't need to
     # track modules, only the compile env. So far this is only true for Elixir's
     # dbg callback.
     compile_env_apps = deps_config_compile_env_apps(old_deps_config)
 
-    # The app tracer will return information about apps before this compilation.
-    app_tracer = Mix.Compilers.ApplicationTracer.init()
-
     {force?, stale, new_deps_config} =
       cond do
         !!opts[:force] or is_nil(old_deps_config) or old_cache_key != new_cache_key ->
-          {true, stale, deps_config()}
+          {true, stale, deps_config(local_deps)}
 
         deps_changed? or compile_env_apps != [] ->
-          new_deps_config = deps_config()
-          config_apps = merge_appset(old_deps_config.config, new_deps_config.config, [])
+          new_deps_config = deps_config(local_deps)
+          local_apps = merge_appset(old_deps_config.local, new_deps_config.local, [])
+          config_apps = merge_appset(old_deps_config.config, new_deps_config.config, local_apps)
           apps = merge_appset(old_deps_config.lock, new_deps_config.lock, config_apps)
 
           if Mix.Project.config()[:app] in apps do
             {true, stale, new_deps_config}
           else
+            app_modules = Mix.AppLoader.read_cache()
+
             apps_stale =
               apps
               |> deps_on()
               |> Enum.flat_map(fn {app, _} ->
                 new_modules = Application.spec(app, :modules) || []
 
-                if old_modules = Mix.Compilers.ApplicationTracer.app_modules(app_tracer, app) do
+                if old_modules = app_modules[app] do
                   :ordsets.union(old_modules, :ordsets.from_list(new_modules))
                 else
                   new_modules
@@ -114,7 +126,7 @@ defmodule Mix.Compilers.Elixir do
       end
 
     {stale_modules, stale_exports, all_local_exports} =
-      stale_local_deps(manifest, stale, modified, all_local_exports)
+      stale_local_deps(local_deps, manifest, stale, modified, all_local_exports)
 
     prev_paths = for source(source: source) <- all_sources, do: source
     removed = prev_paths -- all_paths
@@ -142,30 +154,33 @@ defmodule Mix.Compilers.Elixir do
     {sources, removed_modules} =
       update_stale_sources(sources, stale, removed_modules, sources_stats)
 
-    if stale != [] do
+    if stale != [] or stale_modules != %{} do
+      path = opts[:purge_consolidation_path_if_stale]
+
+      if is_binary(path) and Code.delete_path(path) do
+        purge_modules_in_path(path)
+      end
+
       Mix.Utils.compiling_n(length(stale), :ex)
       Mix.Project.ensure_structure()
+
+      # We don't want to cache this path as we will write to it
       true = Code.prepend_path(dest)
-
-      {pending_tracer, tracer_opts} = Mix.Compilers.ApplicationTracer.prepare(app_tracer, opts)
-      previous_opts = set_compiler_opts(tracer_opts)
-
-      # Stores state for keeping track which files were compiled
-      # and the dependencies between them.
-      put_compiler_info({[], exports, sources, modules, removed_modules})
+      previous_opts = set_compiler_opts(opts)
 
       try do
-        compile_path(stale, dest, timestamp, opts)
+        state = {[], exports, sources, modules, removed_modules}
+        compiler_loop(stale, stale_modules, dest, timestamp, opts, state)
       else
-        {:ok, _, warnings} ->
-          {modules, _exports, sources, pending_modules, _pending_exports} = get_compiler_info()
+        {:ok, info, state} ->
+          {modules, _exports, sources, pending_modules, _pending_exports} = state
 
-          # We only collect the warnings if --all-warnings is given.
-          # In this case, we print them too. Then we apply the new warnings.
           previous_warnings =
-            if opts[:all_warnings], do: previous_warnings(sources, true), else: []
+            if Keyword.get(opts, :all_warnings, true),
+              do: previous_warnings(sources, true),
+              else: []
 
-          sources = apply_warnings(sources, warnings)
+          sources = apply_warnings(sources, info)
 
           write_manifest(
             manifest,
@@ -179,21 +194,20 @@ defmodule Mix.Compilers.Elixir do
           )
 
           put_compile_env(sources)
-          all_warnings = previous_warnings ++ Enum.map(warnings, &diagnostic(&1, :warning))
+          info_warnings = info.runtime_warnings ++ info.compile_warnings
+          all_warnings = previous_warnings ++ Enum.map(info_warnings, &diagnostic/1)
           unless_previous_warnings_as_errors(previous_warnings, opts, {:ok, all_warnings})
 
-        {:error, errors, warnings} ->
+        {:error, errors, %{runtime_warnings: r_warnings, compile_warnings: c_warnings}, state} ->
           # In case of errors, we show all previous warnings and all new ones.
-          # Print the new ones if --all-warnings was given.
-          {_, _, sources, _, _} = get_compiler_info()
-          errors = Enum.map(errors, &diagnostic(&1, :error))
-          warnings = Enum.map(warnings, &diagnostic(&1, :warning))
-          {:error, previous_warnings(sources, opts[:all_warnings]) ++ warnings ++ errors}
+          {_, _, sources, _, _} = state
+          errors = Enum.map(errors, &diagnostic/1)
+          warnings = Enum.map(r_warnings ++ c_warnings, &diagnostic/1)
+          all_warnings = Keyword.get(opts, :all_warnings, true)
+          {:error, previous_warnings(sources, all_warnings) ++ warnings ++ errors}
       after
         Code.compiler_options(previous_opts)
-        Mix.Compilers.ApplicationTracer.stop(pending_tracer)
         Code.purge_compiler_modules()
-        delete_compiler_info()
       end
     else
       # We need to return ok if deps_changed? or stale_modules changed,
@@ -225,22 +239,26 @@ defmodule Mix.Compilers.Elixir do
         )
       end
 
-      previous_warnings = previous_warnings(sources, opts[:all_warnings])
+      all_warnings = Keyword.get(opts, :all_warnings, true)
+      previous_warnings = previous_warnings(sources, all_warnings)
       unless_previous_warnings_as_errors(previous_warnings, opts, {status, previous_warnings})
     end
-  after
-    Mix.Compilers.ApplicationTracer.stop()
   end
 
-  defp deps_config do
+  defp deps_config(local_deps) do
     # If you change this config, you need to bump @manifest_vsn
     %{
+      local: Enum.sort(Enum.map(local_deps, &{&1.app, true})),
       lock: Enum.sort(Mix.Dep.Lock.read()),
       config: Enum.sort(Mix.Tasks.Loadconfig.read_compile()),
       dbg: Application.fetch_env!(:elixir, :dbg_callback)
     }
   end
 
+  defp local_deps_changed?(deps_config, local_deps) do
+    is_map(deps_config) and Enum.sort(Enum.map(local_deps, &{&1.app, true})) != deps_config.local
+  end
+
   defp deps_config_compile_env_apps(deps_config) do
     if deps_config[:dbg] != Application.fetch_env!(:elixir, :dbg_callback) do
       [:elixir]
@@ -320,10 +338,31 @@ defmodule Mix.Compilers.Elixir do
          stale_exports,
          dest
        ) do
+    {modules_to_recompile, modules_to_mix_check} =
+      for module(module: module, recompile?: recompile?) <- all_modules, reduce: {[], []} do
+        {modules_to_recompile, modules_to_mix_check} ->
+          cond do
+            Map.has_key?(stale_modules, module) ->
+              {[module | modules_to_recompile], modules_to_mix_check}
+
+            recompile? and Code.ensure_loaded?(module) and
+                function_exported?(module, :__mix_recompile__?, 0) ->
+              {modules_to_recompile, [module | modules_to_mix_check]}
+
+            true ->
+              {modules_to_recompile, modules_to_mix_check}
+          end
+      end
+
     modules_to_recompile =
-      for module(module: module, recompile?: recompile?) <- all_modules,
-          recompile_module?(module, recompile?) or Map.has_key?(stale_modules, module),
-          do: module
+      modules_to_recompile ++
+        for {:ok, {module, true}} <-
+              Task.async_stream(modules_to_mix_check, &{&1, &1.__mix_recompile__?()},
+                ordered: false,
+                timeout: :infinity
+              ) do
+          module
+        end
 
     {checkpoint_stale_modules, checkpoint_stale_exports, checkpoint_modules} =
       parse_checkpoint(manifest)
@@ -359,7 +398,7 @@ defmodule Mix.Compilers.Elixir do
             Enum.any?(modules, &Map.has_key?(modules_to_recompile, &1)) or
             Enum.any?(external, &stale_external?(&1, modified, sources_stats)) or
             (last_mtime > modified and
-               (missing_beam_file?(dest, modules) or digest != digest(source))),
+               (missing_beam_file?(dest, modules) or digest_changed?(source, digest))),
           do: source
 
     changed = new_paths ++ changed
@@ -386,10 +425,10 @@ defmodule Mix.Compilers.Elixir do
     {modules, exports, changed, sources_stats}
   end
 
-  defp stale_external?({external, existed?}, modified, sources_stats) do
+  defp stale_external?({external, digest}, modified, sources_stats) do
     case sources_stats do
-      %{^external => {0, 0}} -> existed?
-      %{^external => {mtime, _}} -> mtime > modified
+      %{^external => {0, 0}} -> digest != nil
+      %{^external => {mtime, _}} -> mtime > modified and digest_changed?(external, digest)
     end
   end
 
@@ -403,34 +442,23 @@ defmodule Mix.Compilers.Elixir do
     end)
   end
 
-  defp digest(file) do
-    file
-    |> File.read!()
-    |> :erlang.md5()
+  defp digest_changed?(file, digest) do
+    case File.read(file) do
+      {:ok, binary} -> digest != digest_contents(binary)
+      {:error, _} -> true
+    end
   end
 
-  defp compile_path(stale, dest, timestamp, opts) do
-    cwd = File.cwd!()
-    long_compilation_threshold = opts[:long_compilation_threshold] || 10
-    verbose = opts[:verbose] || false
-
-    compile_opts = [
-      each_cycle: fn -> each_cycle(dest, timestamp) end,
-      each_file: &each_file(&1, &2, cwd, verbose),
-      each_module: &each_module(&1, &2, &3, cwd),
-      each_long_compilation: &each_long_compilation(&1, cwd, long_compilation_threshold),
-      long_compilation_threshold: long_compilation_threshold,
-      profile: opts[:profile],
-      beam_timestamp: timestamp
-    ]
-
-    Kernel.ParallelCompiler.compile_to_path(stale, dest, compile_opts)
+  defp digest_contents(contents) do
+    case :erlang.system_info(:wordsize) do
+      8 -> :crypto.hash(:blake2b, contents)
+      _ -> :crypto.hash(:blake2s, contents)
+    end
+  rescue
+    # Blake may not be available on all OpenSSL distribution
+    _ -> :erlang.md5(contents)
   end
 
-  defp get_compiler_info(), do: Process.get(__MODULE__)
-  defp put_compiler_info(value), do: Process.put(__MODULE__, value)
-  defp delete_compiler_info(), do: Process.delete(__MODULE__)
-
   defp set_compiler_opts(opts) do
     opts
     |> Keyword.take(Code.available_compiler_options())
@@ -446,172 +474,6 @@ defmodule Mix.Compilers.Elixir do
     Mix.ProjectStack.compile_env(all_compile_env)
   end
 
-  defp each_cycle(compile_path, timestamp) do
-    {modules, _exports, sources, pending_modules, pending_exports} = get_compiler_info()
-
-    {pending_modules, exports, changed} =
-      update_stale_entries(pending_modules, sources, [], %{}, pending_exports, compile_path)
-
-    # For each changed file, mark it as changed.
-    # If compilation fails mid-cycle, they will
-    # be picked next time around.
-    for file <- changed do
-      File.touch!(file, timestamp)
-    end
-
-    if changed == [] do
-      modules = Enum.map(modules, &module(&1, :module))
-      warnings = Mix.Compilers.ApplicationTracer.warnings(modules)
-
-      modules_set = Map.from_keys(modules, true)
-      {_, runtime_modules} = fixpoint_runtime_modules(sources, modules_set)
-
-      runtime_paths =
-        Enum.map(runtime_modules, &{&1, Path.join(compile_path, Atom.to_string(&1) <> ".beam")})
-
-      {:runtime, runtime_paths, warnings}
-    else
-      Mix.Utils.compiling_n(length(changed), :ex)
-
-      # If we have a compile time dependency to a module, as soon as its file
-      # change, we will detect the compile time dependency and recompile. However,
-      # the whole goal of pending exports is to delay this decision, so we need to
-      # track which modules were removed and start them as our pending exports and
-      # remove the pending exports as we notice they have not gone stale.
-      {sources, removed_modules} = update_stale_sources(sources, changed)
-      put_compiler_info({modules, exports, sources, pending_modules, removed_modules})
-      {:compile, changed, []}
-    end
-  end
-
-  defp each_module(file, module, _binary, cwd) do
-    {modules, exports, sources, pending_modules, pending_exports} = get_compiler_info()
-
-    kind = detect_kind(module)
-    file = Path.relative_to(file, cwd)
-    external = get_external_resources(module, cwd)
-
-    old_export = Map.get(exports, module)
-    new_export = exports_md5(module, true)
-
-    pending_exports =
-      if old_export && old_export != new_export do
-        pending_exports
-      else
-        Map.delete(pending_exports, module)
-      end
-
-    {module_sources, existing_module?} =
-      case List.keyfind(modules, module, module(:module)) do
-        module(sources: old_sources) -> {[file | List.delete(old_sources, file)], true}
-        nil -> {[file], false}
-      end
-
-    {source, sources} =
-      List.keytake(sources, file, source(:source)) ||
-        Mix.raise(
-          "Could not find source for #{inspect(file)}. Make sure the :elixirc_paths configuration " <>
-            "is a list of relative paths to the current project or absolute paths to external directories"
-        )
-
-    source =
-      source(
-        source,
-        external: external ++ source(source, :external),
-        modules: [module | source(source, :modules)]
-      )
-
-    module =
-      module(
-        module: module,
-        kind: kind,
-        sources: module_sources,
-        export: new_export,
-        recompile?: function_exported?(module, :__mix_recompile__?, 0)
-      )
-
-    modules = prepend_or_merge(modules, module, module(:module), module, existing_module?)
-    put_compiler_info({modules, exports, [source | sources], pending_modules, pending_exports})
-    :ok
-  end
-
-  defp recompile_module?(module, recompile?) do
-    recompile? and Code.ensure_loaded?(module) and
-      function_exported?(module, :__mix_recompile__?, 0) and
-      module.__mix_recompile__?()
-  end
-
-  defp prepend_or_merge(collection, key, pos, value, true) do
-    List.keystore(collection, key, pos, value)
-  end
-
-  defp prepend_or_merge(collection, _key, _pos, value, false) do
-    [value | collection]
-  end
-
-  defp detect_kind(module) do
-    protocol_metadata = Module.get_attribute(module, :__impl__)
-
-    cond do
-      is_list(protocol_metadata) and protocol_metadata[:protocol] ->
-        {:impl, protocol_metadata[:protocol]}
-
-      is_list(Module.get_attribute(module, :__protocol__)) ->
-        :protocol
-
-      true ->
-        :module
-    end
-  end
-
-  defp get_external_resources(module, cwd) do
-    for file <- Module.get_attribute(module, :external_resource) do
-      {Path.relative_to(file, cwd), File.exists?(file)}
-    end
-  end
-
-  defp each_file(file, lexical, cwd, verbose) do
-    file = Path.relative_to(file, cwd)
-
-    if verbose do
-      Mix.shell().info("Compiled #{file}")
-    end
-
-    {modules, exports, sources, pending_modules, pending_exports} = get_compiler_info()
-    {source, sources} = List.keytake(sources, file, source(:source))
-
-    {compile_references, export_references, runtime_references, compile_env} =
-      Kernel.LexicalTracker.references(lexical)
-
-    compile_references =
-      Enum.reject(compile_references, &match?("elixir_" <> _, Atom.to_string(&1)))
-
-    source(modules: source_modules) = source
-    compile_references = compile_references -- source_modules
-    export_references = export_references -- source_modules
-    runtime_references = runtime_references -- source_modules
-
-    source =
-      source(
-        source,
-        # We preserve the digest if the file is recompiled but not changed
-        digest: source(source, :digest) || digest(file),
-        compile_references: compile_references,
-        export_references: export_references,
-        runtime_references: runtime_references,
-        compile_env: compile_env
-      )
-
-    put_compiler_info({modules, exports, [source | sources], pending_modules, pending_exports})
-    :ok
-  end
-
-  defp each_long_compilation(file, cwd, threshold) do
-    Mix.shell().info(
-      "Compiling #{Path.relative_to(file, cwd)} (it's taking more than #{threshold}s)"
-    )
-  end
-
   ## Resolution
 
   defp remove_removed_sources(sources, removed) do
@@ -641,18 +503,6 @@ defmodule Mix.Compilers.Elixir do
     end)
   end
 
-  # Define empty records for the sources that needs
-  # to be recompiled (but were not changed on disk)
-  defp update_stale_sources(sources, changed) do
-    Enum.reduce(changed, {sources, %{}}, fn file, {acc_sources, acc_modules} ->
-      {source(size: size, digest: digest, modules: modules), acc_sources} =
-        List.keytake(acc_sources, file, source(:source))
-
-      acc_modules = Enum.reduce(modules, acc_modules, &Map.put(&2, &1, true))
-      {[source(source: file, size: size, digest: digest) | acc_sources], acc_modules}
-    end)
-  end
-
   # This function receives the manifest entries and some source
   # files that have changed. Then it recursively figures out
   # all the files that changed (via the module dependencies) and
@@ -719,36 +569,37 @@ defmodule Mix.Compilers.Elixir do
     Enum.any?(enumerable, &Map.has_key?(map, &1))
   end
 
-  defp stale_local_deps(manifest, stale_modules, modified, old_exports) do
+  defp stale_local_deps(local_deps, manifest, stale_modules, modified, deps_exports) do
     base = Path.basename(manifest)
 
     # The stale modules so far will become both stale_modules and stale_exports,
     # as any export from a dependency needs to be recompiled.
     stale_modules = Map.from_keys(stale_modules, true)
 
-    for %{scm: scm, opts: opts} = dep <- Mix.Dep.cached(),
-        not scm.fetchable?,
+    for %{app: app, opts: opts} <- local_deps,
         manifest = Path.join([opts[:build], ".mix", base]),
         Mix.Utils.last_modified(manifest) > modified,
-        reduce: {stale_modules, stale_modules, old_exports} do
-      {modules, exports, new_exports} ->
-        {_manifest_modules, dep_sources} = read_manifest(manifest)
+        reduce: {stale_modules, stale_modules, deps_exports} do
+      {modules, exports, deps_exports} ->
+        {manifest_modules, manifest_sources} = read_manifest(manifest)
 
         dep_modules =
-          for path <- Mix.Dep.load_paths(dep),
-              beam <- Path.wildcard(Path.join(path, "*.beam")),
-              Mix.Utils.last_modified(beam) > modified,
-              do: beam |> Path.basename() |> Path.rootname() |> String.to_atom()
+          for module(module: module, timestamp: timestamp) <- manifest_modules,
+              timestamp > modified,
+              do: module
 
         # If any module has a compile time dependency on a changed module
-        # within the dependnecy, they will be recompiled. However, export
+        # within the dependency, they will be recompiled. However, export
         # and runtime dependencies won't have recompiled so we need to
         # propagate them to the parent app.
-        {dep_modules, _} = fixpoint_runtime_modules(dep_sources, Map.from_keys(dep_modules, true))
+        {dep_modules, _, _} =
+          fixpoint_runtime_modules(manifest_sources, Map.from_keys(dep_modules, true))
+
+        old_exports = Map.get(deps_exports, app, %{})
 
         # Update exports
         {exports, new_exports} =
-          for {module, _} <- dep_modules, reduce: {exports, new_exports} do
+          for {module, _} <- dep_modules, reduce: {exports, []} do
             {exports, new_exports} ->
               export = exports_md5(module, false)
 
@@ -762,29 +613,46 @@ defmodule Mix.Compilers.Elixir do
                   do: exports,
                   else: Map.put(exports, module, true)
 
-              # In any case, we always store it as the most update export
-              # that we have, otherwise we delete it.
+              # Then we store the new export if any
               new_exports =
                 if export,
-                  do: Map.put(new_exports, module, export),
-                  else: Map.delete(new_exports, module)
+                  do: [{module, export} | new_exports],
+                  else: new_exports
 
               {exports, new_exports}
           end
 
-        {Map.merge(modules, dep_modules), exports, new_exports}
+        new_exports = Map.new(new_exports)
+
+        removed =
+          for {module, _} <- old_exports,
+              not is_map_key(new_exports, module),
+              do: {module, true},
+              into: %{}
+
+        modules = modules |> Map.merge(dep_modules) |> Map.merge(removed)
+        exports = Map.merge(exports, removed)
+        deps_exports = Map.put(deps_exports, app, new_exports)
+        {modules, exports, deps_exports}
     end
   end
 
   defp fixpoint_runtime_modules(sources, modules) when modules != %{} do
-    fixpoint_runtime_modules(sources, modules, false, [], [])
+    fixpoint_runtime_modules(sources, modules, false, [], [], [])
   end
 
-  defp fixpoint_runtime_modules(_sources, modules) do
-    {modules, []}
+  defp fixpoint_runtime_modules(sources, modules) do
+    {modules, [], sources}
   end
 
-  defp fixpoint_runtime_modules([source | sources], modules, new?, acc_modules, acc_sources) do
+  defp fixpoint_runtime_modules(
+         [source | sources],
+         modules,
+         new?,
+         pending_sources,
+         acc_modules,
+         acc_sources
+       ) do
     source(export_references: export_refs, runtime_references: runtime_refs) = source
 
     if has_any_key?(modules, export_refs) or has_any_key?(modules, runtime_refs) do
@@ -792,18 +660,20 @@ defmodule Mix.Compilers.Elixir do
       modules = Enum.reduce(new_modules, modules, &Map.put(&2, &1, true))
       new? = new? or new_modules != []
       acc_modules = new_modules ++ acc_modules
-      fixpoint_runtime_modules(sources, modules, new?, acc_modules, acc_sources)
+      acc_sources = [source(source, runtime_warnings: []) | acc_sources]
+      fixpoint_runtime_modules(sources, modules, new?, pending_sources, acc_modules, acc_sources)
     else
-      fixpoint_runtime_modules(sources, modules, new?, acc_modules, [source | acc_sources])
+      pending_sources = [source | pending_sources]
+      fixpoint_runtime_modules(sources, modules, new?, pending_sources, acc_modules, acc_sources)
     end
   end
 
-  defp fixpoint_runtime_modules([], modules, new?, acc_modules, acc_sources)
-       when new? == false or acc_sources == [],
-       do: {modules, acc_modules}
+  defp fixpoint_runtime_modules([], modules, new?, pending_sources, acc_modules, acc_sources)
+       when new? == false or pending_sources == [],
+       do: {modules, acc_modules, acc_sources ++ pending_sources}
 
-  defp fixpoint_runtime_modules([], modules, true, acc_modules, acc_sources),
-    do: fixpoint_runtime_modules(acc_sources, modules, false, acc_modules, [])
+  defp fixpoint_runtime_modules([], modules, true, pending_sources, acc_modules, acc_sources),
+    do: fixpoint_runtime_modules(pending_sources, modules, false, [], acc_modules, acc_sources)
 
   defp exports_md5(module, use_attributes?) do
     cond do
@@ -829,35 +699,78 @@ defmodule Mix.Compilers.Elixir do
 
   defp remove_and_purge(beam, module) do
     _ = File.rm(beam)
-    _ = :code.purge(module)
-    _ = :code.delete(module)
+    :code.purge(module)
+    :code.delete(module)
+  end
+
+  defp purge_modules_in_path(path) do
+    with {:ok, beams} <- File.ls(path) do
+      Enum.each(beams, fn beam ->
+        module = beam |> Path.rootname() |> String.to_atom()
+        :code.purge(module)
+        :code.delete(module)
+      end)
+    end
   end
 
   defp previous_warnings(sources, print?) do
-    for source(source: source, warnings: warnings) <- sources,
+    for source(
+          source: source,
+          compile_warnings: compile_warnings,
+          runtime_warnings: runtime_warnings
+        ) <- sources,
         file = Path.absname(source),
-        {location, message} <- warnings do
-      warning = {file, location, message}
-      print? && Kernel.ParallelCompiler.print_warning(warning)
-      diagnostic(warning, :warning)
+        {position, message} <- compile_warnings ++ runtime_warnings do
+      diagnostic = %Mix.Task.Compiler.Diagnostic{
+        severity: :warning,
+        file: file,
+        position: position,
+        message: message,
+        compiler_name: "Elixir",
+        stacktrace: []
+      }
+
+      if print? do
+        Mix.shell().print_app()
+        Code.print_diagnostic(diagnostic)
+      end
+
+      diagnostic
     end
   end
 
-  defp apply_warnings(sources, warnings) do
-    warnings = Enum.group_by(warnings, &elem(&1, 0), &{elem(&1, 1), elem(&1, 2)})
+  defp apply_warnings(sources, %{runtime_warnings: r_warnings, compile_warnings: c_warnings}) do
+    runtime_group = Enum.group_by(r_warnings, & &1.file, &{&1.position, &1.message})
+    compile_group = Enum.group_by(c_warnings, & &1.file, &{&1.position, &1.message})
 
-    for source(source: source_path, warnings: source_warnings) = s <- sources do
-      source(s, warnings: Map.get(warnings, Path.absname(source_path), source_warnings))
+    for source(
+          source: source_path,
+          runtime_warnings: runtime_warnings,
+          compile_warnings: compile_warnings
+        ) = s <- sources do
+      key = Path.absname(source_path)
+
+      source(s,
+        runtime_warnings: Map.get(runtime_group, key, runtime_warnings),
+        compile_warnings: Map.get(compile_group, key, compile_warnings)
+      )
     end
   end
 
-  defp diagnostic({file, location, message}, severity) do
+  defp diagnostic(%{
+         file: file,
+         position: position,
+         message: message,
+         severity: severity,
+         stacktrace: stacktrace
+       }) do
     %Mix.Task.Compiler.Diagnostic{
       file: file,
-      position: location,
+      position: position,
       message: message,
       severity: severity,
-      compiler_name: "Elixir"
+      compiler_name: "Elixir",
+      stacktrace: stacktrace
     }
   end
 
@@ -926,8 +839,7 @@ defmodule Mix.Compilers.Elixir do
       {@manifest_vsn, modules, sources, local_exports, parent, cache_key, deps_config} ->
         {modules, sources, local_exports, parent, cache_key, deps_config}
 
-      # {vsn, modules, sources} v5-v7 (v1.10)
-      # {vsn, modules, sources, local_exports} v8-v10 (v1.11)
+      # {vsn, modules, sources, ...} v5-v16
       manifest when is_tuple(manifest) and is_integer(elem(manifest, 0)) ->
         purge_old_manifest(compile_path, elem(manifest, 1))
 
@@ -1037,4 +949,249 @@ defmodule Mix.Compilers.Elixir do
       {status, all_warnings}
     end
   end
+
+  ## Compiler loop
+  # The compiler is invoked in a separate process so we avoid blocking its main loop.
+
+  defp compiler_loop(stale, stale_modules, dest, timestamp, opts, state) do
+    ref = make_ref()
+    parent = self()
+    threshold = opts[:long_compilation_threshold] || 10
+    profile = opts[:profile]
+    verbose = opts[:verbose] || false
+
+    pid =
+      spawn_link(fn ->
+        compile_opts = [
+          each_cycle: fn ->
+            compiler_call(parent, ref, {:each_cycle, stale_modules, dest, timestamp})
+          end,
+          each_file: fn file, lexical ->
+            compiler_call(parent, ref, {:each_file, file, lexical, verbose})
+          end,
+          each_module: fn file, module, _binary ->
+            compiler_call(parent, ref, {:each_module, file, module, System.os_time(:second)})
+          end,
+          each_long_compilation: fn file ->
+            Mix.shell().info(
+              "Compiling #{Path.relative_to(file, File.cwd!())} (it's taking more than #{threshold}s)"
+            )
+          end,
+          long_compilation_threshold: threshold,
+          profile: profile,
+          beam_timestamp: timestamp,
+          return_diagnostics: true
+        ]
+
+        response = Kernel.ParallelCompiler.compile_to_path(stale, dest, compile_opts)
+        send(parent, {ref, response})
+      end)
+
+    compiler_loop(ref, pid, state, File.cwd!())
+  end
+
+  defp compiler_call(parent, ref, info) do
+    send(parent, {ref, info})
+
+    receive do
+      {^ref, response} -> response
+    end
+  end
+
+  defp compiler_loop(ref, pid, state, cwd) do
+    receive do
+      {^ref, {:each_cycle, stale_modules, dest, timestamp}} ->
+        {response, state} = each_cycle(stale_modules, dest, timestamp, state)
+        send(pid, {ref, response})
+        compiler_loop(ref, pid, state, cwd)
+
+      {^ref, {:each_file, file, lexical, verbose}} ->
+        # Read the relevant file information and unblock the compiler
+        references = Kernel.LexicalTracker.references(lexical)
+        send(pid, {ref, :ok})
+        state = each_file(file, references, verbose, state, cwd)
+        compiler_loop(ref, pid, state, cwd)
+
+      {^ref, {:each_module, file, module, timestamp}} ->
+        # Read the relevant module information and unblock the compiler
+        kind = detect_kind(module)
+        external = Module.get_attribute(module, :external_resource)
+        new_export = exports_md5(module, true)
+        send(pid, {ref, :ok})
+        state = each_module(file, module, kind, external, new_export, state, timestamp, cwd)
+        compiler_loop(ref, pid, state, cwd)
+
+      {^ref, {:ok, _modules, info}} ->
+        {:ok, info, state}
+
+      {^ref, {:error, errors, info}} ->
+        {:error, errors, info, state}
+    end
+  end
+
+  defp each_cycle(stale_modules, compile_path, timestamp, state) do
+    {modules, _exports, sources, pending_modules, pending_exports} = state
+
+    {pending_modules, exports, changed} =
+      update_stale_entries(pending_modules, sources, [], %{}, pending_exports, compile_path)
+
+    # For each changed file, mark it as changed.
+    # If compilation fails mid-cycle, they will be picked next time around.
+    for file <- changed do
+      File.touch!(file, timestamp)
+    end
+
+    if changed == [] do
+      stale_modules =
+        modules
+        |> Enum.map(&module(&1, :module))
+        |> Map.from_keys(true)
+        |> Map.merge(stale_modules)
+
+      {_, runtime_modules, sources} = fixpoint_runtime_modules(sources, stale_modules)
+
+      runtime_paths =
+        Enum.map(runtime_modules, &{&1, Path.join(compile_path, Atom.to_string(&1) <> ".beam")})
+
+      state = {modules, exports, sources, pending_modules, pending_exports}
+      {{:runtime, runtime_paths, []}, state}
+    else
+      Mix.Utils.compiling_n(length(changed), :ex)
+
+      # If we have a compile time dependency to a module, as soon as its file
+      # change, we will detect the compile time dependency and recompile. However,
+      # the whole goal of pending exports is to delay this decision, so we need to
+      # track which modules were removed and start them as our pending exports and
+      # remove the pending exports as we notice they have not gone stale.
+      {sources, removed_modules} =
+        Enum.reduce(changed, {sources, %{}}, fn file, {acc_sources, acc_modules} ->
+          {source(size: size, digest: digest, modules: modules), acc_sources} =
+            List.keytake(acc_sources, file, source(:source))
+
+          acc_modules = Enum.reduce(modules, acc_modules, &Map.put(&2, &1, true))
+
+          # Define empty records for the sources that needs
+          # to be recompiled (but were not changed on disk)
+          {[source(source: file, size: size, digest: digest) | acc_sources], acc_modules}
+        end)
+
+      state = {modules, exports, sources, pending_modules, removed_modules}
+      {{:compile, changed, []}, state}
+    end
+  end
+
+  defp each_file(file, references, verbose, state, cwd) do
+    {compile_references, export_references, runtime_references, compile_env} = references
+    {modules, exports, sources, pending_modules, pending_exports} = state
+
+    file = Path.relative_to(file, cwd)
+
+    if verbose do
+      Mix.shell().info("Compiled #{file}")
+    end
+
+    {source, sources} = List.keytake(sources, file, source(:source))
+
+    compile_references =
+      Enum.reject(compile_references, &match?("elixir_" <> _, Atom.to_string(&1)))
+
+    source(modules: source_modules) = source
+    compile_references = compile_references -- source_modules
+    export_references = export_references -- source_modules
+    runtime_references = runtime_references -- source_modules
+
+    source =
+      source(
+        source,
+        # We preserve the digest if the file is recompiled but not changed
+        digest: source(source, :digest) || file |> File.read!() |> digest_contents(),
+        compile_references: compile_references,
+        export_references: export_references,
+        runtime_references: runtime_references,
+        compile_env: compile_env
+      )
+
+    {modules, exports, [source | sources], pending_modules, pending_exports}
+  end
+
+  defp each_module(file, module, kind, external, new_export, state, timestamp, cwd) do
+    {modules, exports, sources, pending_modules, pending_exports} = state
+
+    file = Path.relative_to(file, cwd)
+    external = process_external_resources(external, cwd)
+
+    old_export = Map.get(exports, module)
+
+    pending_exports =
+      if old_export && old_export != new_export do
+        pending_exports
+      else
+        Map.delete(pending_exports, module)
+      end
+
+    {module_sources, existing_module?} =
+      case List.keyfind(modules, module, module(:module)) do
+        module(sources: old_sources) -> {[file | List.delete(old_sources, file)], true}
+        nil -> {[file], false}
+      end
+
+    {source, sources} =
+      List.keytake(sources, file, source(:source)) ||
+        Mix.raise(
+          "Could not find source for #{inspect(file)}. Make sure the :elixirc_paths configuration " <>
+            "is a list of relative paths to the current project or absolute paths to external directories"
+        )
+
+    source =
+      source(
+        source,
+        external: external ++ source(source, :external),
+        modules: [module | source(source, :modules)]
+      )
+
+    module =
+      module(
+        module: module,
+        kind: kind,
+        sources: module_sources,
+        export: new_export,
+        timestamp: timestamp,
+        recompile?: function_exported?(module, :__mix_recompile__?, 0)
+      )
+
+    modules = prepend_or_merge(modules, module, module(:module), module, existing_module?)
+    {modules, exports, [source | sources], pending_modules, pending_exports}
+  end
+
+  defp prepend_or_merge(collection, key, pos, value, true) do
+    List.keystore(collection, key, pos, value)
+  end
+
+  defp prepend_or_merge(collection, _key, _pos, value, false) do
+    [value | collection]
+  end
+
+  defp detect_kind(module) do
+    protocol_metadata = Module.get_attribute(module, :__impl__)
+
+    cond do
+      is_list(protocol_metadata) and protocol_metadata[:protocol] ->
+        {:impl, protocol_metadata[:protocol]}
+
+      is_list(Module.get_attribute(module, :__protocol__)) ->
+        :protocol
+
+      true ->
+        :module
+    end
+  end
+
+  defp process_external_resources(external, cwd) do
+    for file <- external do
+      case File.read(file) do
+        {:ok, binary} -> {Path.relative_to(file, cwd), digest_contents(binary)}
+        {:error, _} -> {Path.relative_to(file, cwd), nil}
+      end
+    end
+  end
 end
