@@ -6,7 +6,7 @@ defmodule Mix.Tasks.Format do
   @moduledoc """
   Formats the given files and patterns.
 
-      mix format mix.exs "lib/**/*.{ex,exs}" "test/**/*.{ex,exs}"
+      $ mix format mix.exs "lib/**/*.{ex,exs}" "test/**/*.{ex,exs}"
 
   If any of the files is `-`, then the input is read from stdin and the output
   is written to stdout.
@@ -60,6 +60,10 @@ defmodule Mix.Tasks.Format do
       that the formatted output may differ between Elixir versions as
       improvements and fixes are applied to the formatter.
 
+    * `--no-exit` - only valid when used with `--check-formatted`.
+      Pass this if you don't want this Mix task to fail (and return a non-zero exit code),
+      but still want to check for format errors and print them to the console.
+
     * `--dry-run` - does not save files after formatting.
 
     * `--dot-formatter` - path to the file with formatter configuration.
@@ -78,7 +82,7 @@ defmodule Mix.Tasks.Format do
   such option is not available in your editor of choice, adding the required
   integration is usually a matter of invoking:
 
-      cd $project && mix format $file
+      $ cd $project && mix format $file
 
   where `$file` refers to the current file and `$project` is the root of your
   project.
@@ -117,10 +121,10 @@ defmodule Mix.Tasks.Format do
 
   Now any application can use your formatter as follows:
 
-      # .formatters.exs
+      # .formatter.exs
       [
         # Define the desired plugins
-        plugins: [MixMarkdownFormatter],
+        plugins: [MixMarkdownFormatter, AnotherMarkdownFormatter],
         # Remember to update the inputs list to include the new extensions
         inputs: ["{mix,.formatter}.exs", "{config,lib,test}/**/*.{ex,exs}", "posts/*.{md,markdown}"]
       ]
@@ -129,6 +133,10 @@ defmodule Mix.Tasks.Format do
   sure that your dependencies and your application have been compiled,
   so the relevant plugin code can be loaded. Otherwise a warning is logged.
 
+  In addition, the order by which you input your plugins is the format order.
+  So, in the above `.formatter.exs`, the `MixMarkdownFormatter` will format
+  the markdown files and sigils before `AnotherMarkdownFormatter`.
+
   ## Importing dependencies configuration
 
   This task supports importing formatter configuration from dependencies.
@@ -170,6 +178,7 @@ defmodule Mix.Tasks.Format do
   @switches [
     check_equivalent: :boolean,
     check_formatted: :boolean,
+    no_exit: :boolean,
     dot_formatter: :string,
     dry_run: :boolean,
     stdin_filename: :string
@@ -178,6 +187,25 @@ defmodule Mix.Tasks.Format do
   @manifest "cached_dot_formatter"
   @manifest_vsn 2
 
+  @newline "\n"
+  @blank " "
+
+  @separator "|"
+  @cr "↵"
+  @line_num_pad @blank
+
+  @gutter [
+    del: " -",
+    eq: "  ",
+    ins: " +",
+    skip: "  "
+  ]
+
+  @colors [
+    del: [text: :red, space: :red_background],
+    ins: [text: :green, space: :green_background]
+  ]
+
   @doc """
   Returns which features this plugin should plug into.
   """
@@ -190,82 +218,35 @@ defmodule Mix.Tasks.Format do
 
   @impl true
   def run(args) do
+    cwd = File.cwd!()
     {opts, args} = OptionParser.parse!(args, strict: @switches)
-    {dot_formatter, formatter_opts} = eval_dot_formatter(opts)
+    {dot_formatter, formatter_opts} = eval_dot_formatter(cwd, opts)
 
     if opts[:check_equivalent] do
       IO.warn("--check-equivalent has been deprecated and has no effect")
     end
 
+    if opts[:no_exit] && !opts[:check_formatted] do
+      Mix.raise("--no-exit can only be used together with --check-formatted")
+    end
+
     {formatter_opts_and_subs, _sources} =
-      eval_deps_and_subdirectories(dot_formatter, [], formatter_opts, [dot_formatter])
+      eval_deps_and_subdirectories(cwd, dot_formatter, formatter_opts, [dot_formatter])
+
+    formatter_opts_and_subs = load_plugins(formatter_opts_and_subs)
 
     args
-    |> expand_args(dot_formatter, formatter_opts_and_subs, opts)
+    |> expand_args(cwd, dot_formatter, formatter_opts_and_subs, opts)
     |> Task.async_stream(&format_file(&1, opts), ordered: false, timeout: :infinity)
     |> Enum.reduce({[], []}, &collect_status/2)
-    |> check!()
+    |> check!(opts)
   end
 
-  @doc """
-  Returns a formatter function and the formatter options to
-  be used for the given file.
-
-  The function must be called with the contents of the file
-  to be formatted. The options are returned for reflection
-  purposes.
-  """
-  @doc since: "1.13.0"
-  def formatter_for_file(file, opts \\ []) do
-    {dot_formatter, formatter_opts} = eval_dot_formatter(opts)
-
-    {formatter_opts_and_subs, _sources} =
-      eval_deps_and_subdirectories(dot_formatter, [], formatter_opts, [dot_formatter])
-
-    find_formatter_and_opts_for_file(file, formatter_opts_and_subs)
-  end
-
-  @doc """
-  Returns formatter options to be used for the given file.
-  """
-  # TODO: Deprecate on Elixir v1.17
-  @doc deprecated: "Use formatter_for_file/2 instead"
-  def formatter_opts_for_file(file, opts \\ []) do
-    {_, formatter_opts} = formatter_for_file(file, opts)
-    formatter_opts
-  end
-
-  defp eval_dot_formatter(opts) do
-    cond do
-      dot_formatter = opts[:dot_formatter] ->
-        {dot_formatter, eval_file_with_keyword_list(dot_formatter)}
-
-      File.regular?(".formatter.exs") ->
-        {".formatter.exs", eval_file_with_keyword_list(".formatter.exs")}
-
-      true ->
-        {".formatter.exs", []}
-    end
-  end
-
-  # This function reads exported configuration from the imported
-  # dependencies and subdirectories and deals with caching the result
-  # of reading such configuration in a manifest file.
-  defp eval_deps_and_subdirectories(dot_formatter, prefix, formatter_opts, sources) do
-    deps = Keyword.get(formatter_opts, :import_deps, [])
-    subs = Keyword.get(formatter_opts, :subdirectories, [])
+  defp load_plugins({formatter_opts, subs}) do
     plugins = Keyword.get(formatter_opts, :plugins, [])
 
-    if not is_list(deps) do
-      Mix.raise("Expected :import_deps to return a list of dependencies, got: #{inspect(deps)}")
-    end
-
-    if not is_list(subs) do
-      Mix.raise("Expected :subdirectories to return a list of directories, got: #{inspect(subs)}")
-    end
-
     if not is_list(plugins) do
-      Mix.raise("Expected :plugins to return a list of modules, got: #{inspect(plugins)}")
+      Mix.raise("Expected :plugins to return a list of directories, got: #{inspect(plugins)}")
     end
 
     if plugins != [] do
@@ -292,12 +273,83 @@ defmodule Mix.Tasks.Format do
     sigils =
       for plugin <- plugins,
           sigil <- find_sigils_from_plugins(plugin, formatter_opts),
-          do: {sigil, &plugin.format(&1, &2 ++ formatter_opts)}
+          do: {sigil, plugin}
 
-    formatter_opts =
-      formatter_opts
-      |> Keyword.put(:plugins, plugins)
-      |> Keyword.put(:sigils, sigils)
+    sigils =
+      sigils
+      |> Enum.group_by(&elem(&1, 0), &elem(&1, 1))
+      |> Enum.map(fn {sigil, plugins} ->
+        {sigil,
+         fn input, opts ->
+           Enum.reduce(plugins, input, fn plugin, input ->
+             plugin.format(input, opts ++ formatter_opts)
+           end)
+         end}
+      end)
+
+    {Keyword.put(formatter_opts, :sigils, sigils),
+     Enum.map(subs, fn {path, opts} -> {path, load_plugins(opts)} end)}
+  end
+
+  @doc """
+  Returns a formatter function and the formatter options to
+  be used for the given file.
+
+  The function must be called with the contents of the file
+  to be formatted. The options are returned for reflection
+  purposes.
+  """
+  @doc since: "1.13.0"
+  def formatter_for_file(file, opts \\ []) do
+    cwd = Keyword.get_lazy(opts, :root, &File.cwd!/0)
+    {dot_formatter, formatter_opts} = eval_dot_formatter(cwd, opts)
+
+    {formatter_opts_and_subs, _sources} =
+      eval_deps_and_subdirectories(cwd, dot_formatter, formatter_opts, [dot_formatter])
+
+    formatter_opts_and_subs = load_plugins(formatter_opts_and_subs)
+
+    find_formatter_and_opts_for_file(Path.expand(file, cwd), formatter_opts_and_subs)
+  end
+
+  @doc """
+  Returns formatter options to be used for the given file.
+  """
+  # TODO: Deprecate on Elixir v1.17
+  @doc deprecated: "Use formatter_for_file/2 instead"
+  def formatter_opts_for_file(file, opts \\ []) do
+    {_, formatter_opts} = formatter_for_file(file, opts)
+    formatter_opts
+  end
+
+  defp eval_dot_formatter(cwd, opts) do
+    cond do
+      dot_formatter = opts[:dot_formatter] ->
+        {dot_formatter, eval_file_with_keyword_list(dot_formatter)}
+
+      File.regular?(Path.join(cwd, ".formatter.exs")) ->
+        dot_formatter = Path.join(cwd, ".formatter.exs")
+        {".formatter.exs", eval_file_with_keyword_list(dot_formatter)}
+
+      true ->
+        {".formatter.exs", []}
+    end
+  end
+
+  # This function reads exported configuration from the imported
+  # dependencies and subdirectories and deals with caching the result
+  # of reading such configuration in a manifest file.
+  defp eval_deps_and_subdirectories(cwd, dot_formatter, formatter_opts, sources) do
+    deps = Keyword.get(formatter_opts, :import_deps, [])
+    subs = Keyword.get(formatter_opts, :subdirectories, [])
+
+    if not is_list(deps) do
+      Mix.raise("Expected :import_deps to return a list of dependencies, got: #{inspect(deps)}")
+    end
+
+    if not is_list(subs) do
+      Mix.raise("Expected :subdirectories to return a list of directories, got: #{inspect(subs)}")
+    end
 
     if deps == [] and subs == [] do
       {{formatter_opts, []}, sources}
@@ -306,7 +358,7 @@ defmodule Mix.Tasks.Format do
 
       {{locals_without_parens, subdirectories}, sources} =
         maybe_cache_in_manifest(dot_formatter, manifest, fn ->
-          {subdirectories, sources} = eval_subs_opts(subs, prefix, sources)
+          {subdirectories, sources} = eval_subs_opts(subs, cwd, sources)
           {{eval_deps_opts(deps), subdirectories}, sources}
         end)
 
@@ -370,11 +422,11 @@ defmodule Mix.Tasks.Format do
         do: parenless_call
   end
 
-  defp eval_subs_opts(subs, prefix, sources) do
+  defp eval_subs_opts(subs, cwd, sources) do
     {subs, sources} =
       Enum.flat_map_reduce(subs, sources, fn sub, sources ->
-        prefix = Path.join(prefix ++ [sub])
-        {Path.wildcard(prefix), [Path.join(prefix, ".formatter.exs") | sources]}
+        cwd = Path.expand(sub, cwd)
+        {Path.wildcard(cwd), [Path.join(cwd, ".formatter.exs") | sources]}
       end)
 
     Enum.flat_map_reduce(subs, sources, fn sub, sources ->
@@ -384,7 +436,7 @@ defmodule Mix.Tasks.Format do
         formatter_opts = eval_file_with_keyword_list(sub_formatter)
 
         {formatter_opts_and_subs, sources} =
-          eval_deps_and_subdirectories(:in_memory, [sub], formatter_opts, sources)
+          eval_deps_and_subdirectories(sub, :in_memory, formatter_opts, sources)
 
         {[{sub, formatter_opts_and_subs}], sources}
       else
@@ -394,21 +446,15 @@ defmodule Mix.Tasks.Format do
   end
 
   defp assert_valid_dep_and_fetch_path(dep, deps_paths) when is_atom(dep) do
-    case Map.fetch(deps_paths, dep) do
-      {:ok, path} ->
-        if File.dir?(path) do
-          path
-        else
-          Mix.raise(
-            "Unavailable dependency #{inspect(dep)} given to :import_deps in the formatter configuration. " <>
-              "The dependency cannot be found in the file system, please run \"mix deps.get\" and try again"
-          )
-        end
-
-      :error ->
+    with %{^dep => path} <- deps_paths,
+         true <- File.dir?(path) do
+      path
+    else
+      _ ->
         Mix.raise(
           "Unknown dependency #{inspect(dep)} given to :import_deps in the formatter configuration. " <>
-            "The dependency is not listed in your mix.exs for environment #{inspect(Mix.env())}"
+            "Make sure the dependency is listed in your mix.exs for environment #{inspect(Mix.env())} " <>
+            "and you have run \"mix deps.get\""
         )
     end
   end
@@ -427,7 +473,7 @@ defmodule Mix.Tasks.Format do
     opts
   end
 
-  defp expand_args([], dot_formatter, formatter_opts_and_subs, _opts) do
+  defp expand_args([], cwd, dot_formatter, formatter_opts_and_subs, _opts) do
     if no_entries_in_formatter_opts?(formatter_opts_and_subs) do
       Mix.raise(
         "Expected one or more files/patterns to be given to mix format " <>
@@ -436,13 +482,13 @@ defmodule Mix.Tasks.Format do
     end
 
     dot_formatter
-    |> expand_dot_inputs([], formatter_opts_and_subs, %{})
+    |> expand_dot_inputs(cwd, formatter_opts_and_subs, %{})
     |> Enum.map(fn {file, {_dot_formatter, formatter_opts}} ->
       {file, find_formatter_for_file(file, formatter_opts)}
     end)
   end
 
-  defp expand_args(files_and_patterns, _dot_formatter, {formatter_opts, subs}, opts) do
+  defp expand_args(files_and_patterns, cwd, _dot_formatter, {formatter_opts, subs}, opts) do
     files =
       for file_or_pattern <- files_and_patterns,
           file <- stdin_or_wildcard(file_or_pattern),
@@ -458,7 +504,7 @@ defmodule Mix.Tasks.Format do
 
     for file <- files do
       if file == :stdin do
-        stdin_filename = Keyword.get(opts, :stdin_filename, "stdin.exs")
+        stdin_filename = Path.expand(Keyword.get(opts, :stdin_filename, "stdin.exs"), cwd)
 
         {formatter, _opts} =
           find_formatter_and_opts_for_file(stdin_filename, {formatter_opts, subs})
@@ -471,25 +517,24 @@ defmodule Mix.Tasks.Format do
     end
   end
 
-  defp expand_dot_inputs(dot_formatter, prefix, {formatter_opts, subs}, acc) do
+  defp expand_dot_inputs(dot_formatter, cwd, {formatter_opts, subs}, acc) do
     if no_entries_in_formatter_opts?({formatter_opts, subs}) do
       Mix.raise("Expected :inputs or :subdirectories key in #{dot_formatter}")
     end
 
     map =
       for input <- List.wrap(formatter_opts[:inputs]),
-          file <- Path.wildcard(Path.join(prefix ++ [input]), match_dot: true),
-          do: {expand_relative_to_cwd(file), {dot_formatter, formatter_opts}},
+          file <- Path.wildcard(Path.expand(input, cwd), match_dot: true),
+          do: {file, {dot_formatter, formatter_opts}},
           into: %{}
 
     acc =
       Map.merge(acc, map, fn file, {dot_formatter1, _}, {dot_formatter2, formatter_opts} ->
         Mix.shell().error(
-          "Both #{dot_formatter1} and #{dot_formatter2} specify the file " <>
-            "#{Path.relative_to_cwd(file)} in their :inputs option. To resolve the " <>
-            "conflict, the configuration in #{dot_formatter1} will be ignored. " <>
-            "Please change the list of :inputs in one of the formatter files so only " <>
-            "one of them matches #{Path.relative_to_cwd(file)}"
+          "Both #{dot_formatter1} and #{dot_formatter2} specify the file #{file} in their " <>
+            ":inputs option. To resolve the conflict, the configuration in #{dot_formatter1} " <>
+            "will be ignored. Please change the list of :inputs in one of the formatter files " <>
+            "so only one of them matches #{file}"
         )
 
         {dot_formatter2, formatter_opts}
@@ -497,23 +542,20 @@ defmodule Mix.Tasks.Format do
 
     Enum.reduce(subs, acc, fn {sub, formatter_opts_and_subs}, acc ->
       sub_formatter = Path.join(sub, ".formatter.exs")
-      expand_dot_inputs(sub_formatter, [sub], formatter_opts_and_subs, acc)
+      expand_dot_inputs(sub_formatter, sub, formatter_opts_and_subs, acc)
     end)
   end
 
-  defp expand_relative_to_cwd(path) do
-    case File.cwd() do
-      {:ok, cwd} -> Path.expand(path, cwd)
-      _ -> path
-    end
-  end
-
   defp find_formatter_for_file(file, formatter_opts) do
     ext = Path.extname(file)
 
     cond do
-      plugin = find_plugin_for_extension(formatter_opts, ext) ->
-        &plugin.format(&1, [extension: ext, file: file] ++ formatter_opts)
+      plugins = find_plugins_for_extension(formatter_opts, ext) ->
+        fn input ->
+          Enum.reduce(plugins, input, fn plugin, input ->
+            plugin.format(input, [extension: ext, file: file] ++ formatter_opts)
+          end)
+        end
 
       ext in ~w(.ex .exs) ->
         &elixir_format(&1, [file: file] ++ formatter_opts)
@@ -523,25 +565,27 @@ defmodule Mix.Tasks.Format do
     end
   end
 
-  defp find_plugin_for_extension(formatter_opts, ext) do
+  defp find_plugins_for_extension(formatter_opts, ext) do
     plugins = Keyword.get(formatter_opts, :plugins, [])
 
-    Enum.find(plugins, fn plugin ->
-      Code.ensure_loaded?(plugin) and function_exported?(plugin, :features, 1) and
-        ext in List.wrap(plugin.features(formatter_opts)[:extensions])
-    end)
+    plugins =
+      Enum.filter(plugins, fn plugin ->
+        Code.ensure_loaded?(plugin) and function_exported?(plugin, :features, 1) and
+          ext in List.wrap(plugin.features(formatter_opts)[:extensions])
+      end)
+
+    if plugins != [], do: plugins, else: nil
   end
 
   defp find_formatter_and_opts_for_file(file, formatter_opts_and_subs) do
-    split = file |> Path.relative_to_cwd() |> Path.split()
-    formatter_opts = recur_formatter_opts_for_file(split, formatter_opts_and_subs)
+    formatter_opts = recur_formatter_opts_for_file(file, formatter_opts_and_subs)
     {find_formatter_for_file(file, formatter_opts), formatter_opts}
   end
 
-  defp recur_formatter_opts_for_file(split, {formatter_opts, subs}) do
+  defp recur_formatter_opts_for_file(file, {formatter_opts, subs}) do
     Enum.find_value(subs, formatter_opts, fn {sub, formatter_opts_and_subs} ->
-      if List.starts_with?(split, Path.split(sub)) do
-        recur_formatter_opts_for_file(split, formatter_opts_and_subs)
+      if String.starts_with?(file, sub) do
+        recur_formatter_opts_for_file(file, formatter_opts_and_subs)
       end
     end)
   end
@@ -581,7 +625,7 @@ defmodule Mix.Tasks.Format do
 
     cond do
       check_formatted? ->
-        if input == output, do: :ok, else: {:not_formatted, file}
+        if input == output, do: :ok, else: {:not_formatted, {file, input, output}}
 
       dry_run? ->
         :ok
@@ -614,30 +658,297 @@ defmodule Mix.Tasks.Format do
     {exits, [file | not_formatted]}
   end
 
-  defp check!({[], []}) do
+  defp check!({[], []}, _task_opts) do
     :ok
   end
 
-  defp check!({[{:exit, :stdin, exception, stacktrace} | _], _not_formatted}) do
+  defp check!({[{:exit, :stdin, exception, stacktrace} | _], _not_formatted}, _task_opts) do
     Mix.shell().error("mix format failed for stdin")
     reraise exception, stacktrace
   end
 
-  defp check!({[{:exit, file, exception, stacktrace} | _], _not_formatted}) do
+  defp check!({[{:exit, file, exception, stacktrace} | _], _not_formatted}, _task_opts) do
     Mix.shell().error("mix format failed for file: #{Path.relative_to_cwd(file)}")
     reraise exception, stacktrace
   end
 
-  defp check!({_exits, [_ | _] = not_formatted}) do
-    Mix.raise("""
-    mix format failed due to --check-formatted.
+  defp check!({_exits, [_ | _] = not_formatted}, task_opts) do
+    no_exit? = Keyword.get(task_opts, :no_exit, false)
+
+    message = """
     The following files are not formatted:
 
-    #{to_bullet_list(not_formatted)}
-    """)
+    #{to_diffs(not_formatted)}
+    """
+
+    if no_exit? do
+      Mix.shell().info(message)
+    else
+      Mix.raise("""
+      mix format failed due to --check-formatted.
+      #{message}
+      """)
+    end
   end
 
-  defp to_bullet_list(files) do
-    Enum.map_join(files, "\n", &"  * #{&1 |> to_string() |> Path.relative_to_cwd()}")
+  defp to_diffs(files) do
+    Enum.map_intersperse(files, "\n", fn
+      {:stdin, unformatted, formatted} ->
+        [IO.ANSI.reset(), text_diff_format(unformatted, formatted)]
+
+      {file, unformatted, formatted} ->
+        [
+          IO.ANSI.bright(),
+          IO.ANSI.red(),
+          file,
+          "\n",
+          IO.ANSI.reset(),
+          "\n",
+          text_diff_format(unformatted, formatted)
+        ]
+    end)
+  end
+
+  @doc false
+  @spec text_diff_format(String.t(), String.t()) :: iolist()
+  def text_diff_format(old, new, opts \\ [])
+
+  def text_diff_format(code, code, _opts), do: []
+
+  def text_diff_format(old, new, opts) do
+    opts = Keyword.validate!(opts, after: 2, before: 2, color: IO.ANSI.enabled?(), line: 1)
+    crs? = String.contains?(old, "\r") || String.contains?(new, "\r")
+
+    old = String.split(old, "\n")
+    new = String.split(new, "\n")
+
+    max = max(length(new), length(old))
+    line_num_digits = max |> Integer.digits() |> length()
+    opts = Keyword.put(opts, :line_num_digits, line_num_digits)
+
+    {line, opts} = Keyword.pop!(opts, :line)
+
+    old
+    |> List.myers_difference(new)
+    |> insert_cr_symbols(crs?)
+    |> diff_to_iodata({line, line}, opts)
+  end
+
+  defp diff_to_iodata(diff, line_nums, opts, iodata \\ [])
+
+  defp diff_to_iodata([], _line_nums, _opts, iodata), do: Enum.reverse(iodata)
+
+  defp diff_to_iodata([{:eq, [""]}], _line_nums, _opts, iodata), do: Enum.reverse(iodata)
+
+  defp diff_to_iodata([{:eq, lines}], line_nums, opts, iodata) do
+    lines_after = Enum.take(lines, opts[:after])
+    iodata = lines(iodata, {:eq, lines_after}, line_nums, opts)
+
+    iodata =
+      case length(lines) > opts[:after] do
+        false -> iodata
+        true -> lines(iodata, :skip, opts)
+      end
+
+    Enum.reverse(iodata)
+  end
+
+  defp diff_to_iodata([{:eq, lines} | diff], {line, line}, opts, [] = iodata) do
+    {start, lines_before} = Enum.split(lines, opts[:before] * -1)
+
+    iodata =
+      case length(lines) > opts[:before] do
+        false -> iodata
+        true -> lines(iodata, :skip, opts)
+      end
+
+    line = line + length(start)
+    iodata = lines(iodata, {:eq, lines_before}, {line, line}, opts)
+
+    line = line + length(lines_before)
+    diff_to_iodata(diff, {line, line}, opts, iodata)
+  end
+
+  defp diff_to_iodata([{:eq, lines} | diff], line_nums, opts, iodata) do
+    case length(lines) > opts[:after] + opts[:before] do
+      true ->
+        {lines1, lines2, lines3} = split(lines, opts[:after], opts[:before] * -1)
+
+        iodata =
+          iodata
+          |> lines({:eq, lines1}, line_nums, opts)
+          |> lines(:skip, opts)
+          |> lines({:eq, lines3}, add_line_nums(line_nums, length(lines1) + length(lines2)), opts)
+
+        line_nums = add_line_nums(line_nums, length(lines))
+
+        diff_to_iodata(diff, line_nums, opts, iodata)
+
+      false ->
+        iodata = lines(iodata, {:eq, lines}, line_nums, opts)
+        line_nums = add_line_nums(line_nums, length(lines))
+
+        diff_to_iodata(diff, line_nums, opts, iodata)
+    end
+  end
+
+  defp diff_to_iodata([{:del, [del]}, {:ins, [ins]} | diff], line_nums, opts, iodata) do
+    iodata = lines(iodata, {:chg, del, ins}, line_nums, opts)
+    diff_to_iodata(diff, add_line_nums(line_nums, 1), opts, iodata)
+  end
+
+  defp diff_to_iodata([{kind, lines} | diff], line_nums, opts, iodata) do
+    iodata = lines(iodata, {kind, lines}, line_nums, opts)
+    line_nums = add_line_nums(line_nums, length(lines), kind)
+
+    diff_to_iodata(diff, line_nums, opts, iodata)
+  end
+
+  defp split(list, count1, count2) do
+    {split1, split2} = Enum.split(list, count1)
+    {split2, split3} = Enum.split(split2, count2)
+    {split1, split2, split3}
+  end
+
+  defp lines(iodata, :skip, opts) do
+    line_num = String.duplicate(@blank, opts[:line_num_digits] * 2 + 1)
+    [[line_num, @gutter[:skip], @separator, @newline] | iodata]
+  end
+
+  defp lines(iodata, {:chg, del, ins}, line_nums, opts) do
+    {del, ins} = line_diff(del, ins, opts)
+
+    [
+      [gutter(line_nums, :ins, opts), ins, @newline],
+      [gutter(line_nums, :del, opts), del, @newline]
+      | iodata
+    ]
+  end
+
+  defp lines(iodata, {kind, lines}, line_nums, opts) do
+    lines
+    |> Enum.with_index()
+    |> Enum.reduce(iodata, fn {line, offset}, iodata ->
+      line_nums = add_line_nums(line_nums, offset, kind)
+      [[gutter(line_nums, kind, opts), colorize(line, kind, false, opts), @newline] | iodata]
+    end)
+  end
+
+  defp gutter(line_nums, kind, opts) do
+    [line_num(line_nums, kind, opts), colorize(@gutter[kind], kind, false, opts), @separator]
+  end
+
+  defp line_num({line_num_old, line_num_new}, :eq, opts) do
+    old =
+      line_num_old
+      |> to_string()
+      |> String.pad_leading(opts[:line_num_digits], @line_num_pad)
+
+    new =
+      line_num_new
+      |> to_string()
+      |> String.pad_leading(opts[:line_num_digits], @line_num_pad)
+
+    [old, @blank, new]
+  end
+
+  defp line_num({line_num_old, _line_num_new}, :del, opts) do
+    old =
+      line_num_old
+      |> to_string()
+      |> String.pad_leading(opts[:line_num_digits], @line_num_pad)
+
+    new = String.duplicate(@blank, opts[:line_num_digits])
+    [old, @blank, new]
+  end
+
+  defp line_num({_line_num_old, line_num_new}, :ins, opts) do
+    old = String.duplicate(@blank, opts[:line_num_digits])
+
+    new =
+      line_num_new
+      |> to_string()
+      |> String.pad_leading(opts[:line_num_digits], @line_num_pad)
+
+    [old, @blank, new]
+  end
+
+  defp line_diff(del, ins, opts) do
+    diff = String.myers_difference(del, ins)
+
+    Enum.reduce(diff, {[], []}, fn
+      {:eq, str}, {del, ins} -> {[del | str], [ins | str]}
+      {:del, str}, {del, ins} -> {[del | colorize(str, :del, true, opts)], ins}
+      {:ins, str}, {del, ins} -> {del, [ins | colorize(str, :ins, true, opts)]}
+    end)
+  end
+
+  defp colorize(str, kind, space?, opts) do
+    if Keyword.fetch!(opts, :color) && Keyword.has_key?(@colors, kind) do
+      color = Keyword.fetch!(@colors, kind)
+
+      if space? do
+        str
+        |> String.split(~r/[\t\s]+/, include_captures: true)
+        |> Enum.map(fn
+          <<start::binary-size(1), _::binary>> = str when start in ["\t", "\s"] ->
+            IO.ANSI.format([color[:space], str])
+
+          str ->
+            IO.ANSI.format([color[:text], str])
+        end)
+      else
+        IO.ANSI.format([color[:text], str])
+      end
+    else
+      str
+    end
+  end
+
+  defp add_line_nums({line_num_old, line_num_new}, lines, kind \\ :eq) do
+    case kind do
+      :eq -> {line_num_old + lines, line_num_new + lines}
+      :ins -> {line_num_old, line_num_new + lines}
+      :del -> {line_num_old + lines, line_num_new}
+    end
+  end
+
+  defp insert_cr_symbols(diffs, false), do: diffs
+  defp insert_cr_symbols(diffs, true), do: do_insert_cr_symbols(diffs, [])
+
+  defp do_insert_cr_symbols([], acc), do: Enum.reverse(acc)
+
+  defp do_insert_cr_symbols([{:del, del}, {:ins, ins} | rest], acc) do
+    {del, ins} = do_insert_cr_symbols(del, ins, {[], []})
+    do_insert_cr_symbols(rest, [{:ins, ins}, {:del, del} | acc])
+  end
+
+  defp do_insert_cr_symbols([diff | rest], acc) do
+    do_insert_cr_symbols(rest, [diff | acc])
+  end
+
+  defp do_insert_cr_symbols([left | left_rest], [right | right_rest], {left_acc, right_acc}) do
+    {left, right} = insert_cr_symbol(left, right)
+    do_insert_cr_symbols(left_rest, right_rest, {[left | left_acc], [right | right_acc]})
+  end
+
+  defp do_insert_cr_symbols([], right, {left_acc, right_acc}) do
+    left = Enum.reverse(left_acc)
+    right = right_acc |> Enum.reverse() |> Enum.concat(right)
+    {left, right}
+  end
+
+  defp do_insert_cr_symbols(left, [], {left_acc, right_acc}) do
+    left = left_acc |> Enum.reverse() |> Enum.concat(left)
+    right = Enum.reverse(right_acc)
+    {left, right}
+  end
+
+  defp insert_cr_symbol(left, right) do
+    case {String.ends_with?(left, "\r"), String.ends_with?(right, "\r")} do
+      {bool, bool} -> {left, right}
+      {true, false} -> {String.replace(left, "\r", @cr), right}
+      {false, true} -> {left, String.replace(right, "\r", @cr)}
+    end
   end
 end
