@@ -7,7 +7,7 @@ defmodule Mix.Utils do
   @doc """
   Gets the Mix home.
 
-  It uses the the locations `MIX_HOME`, `XDG_DATA_HOME/mix`,
+  It uses the locations `MIX_HOME`, `XDG_DATA_HOME/mix`,
   `~/.mix` with decreasing priority.
 
   Developers should only store entries in the
@@ -225,6 +225,7 @@ defmodule Mix.Utils do
   @doc """
   Prints n files are being compiled with the given extension.
   """
+  def compiling_n(0, _ext), do: :ok
   def compiling_n(1, ext), do: Mix.shell().info("Compiling 1 file (.#{ext})")
   def compiling_n(n, ext), do: Mix.shell().info("Compiling #{n} files (.#{ext})")
 
@@ -327,6 +328,8 @@ defmodule Mix.Utils do
 
   The callback will be invoked for each node and it
   must return a `{printed, children}` tuple.
+
+  If `path` is `-`, prints the output to standard output.
   """
   @spec write_dot_graph!(
           Path.t(),
@@ -338,7 +341,13 @@ defmodule Mix.Utils do
         when node: term()
   def write_dot_graph!(path, title, nodes, callback, _opts \\ []) do
     {dot, _} = build_dot_graph(make_ref(), nodes, MapSet.new(), callback)
-    File.write!(path, ["digraph ", quoted(title), " {\n", dot, "}\n"])
+    contents = ["digraph ", quoted(title), " {\n", dot, "}\n"]
+
+    if path == "-" do
+      IO.write(contents)
+    else
+      File.write!(path, contents)
+    end
   end
 
   defp build_dot_graph(_parent, [], seen, _callback), do: {[], seen}
@@ -435,10 +444,7 @@ defmodule Mix.Utils do
     |> Enum.map_join(".", &Macro.camelize/1)
   end
 
-  @doc """
-  Symlinks or copy with the option to force a hard copy.
-  See `symlink_or_copy/2`.
-  """
+  @deprecated "Use symlink_or_copy/2"
   def symlink_or_copy(hard_copy?, source, target) do
     if hard_copy? do
       if File.exists?(source) do
@@ -460,6 +466,10 @@ defmodule Mix.Utils do
   Expects source and target to be absolute paths as it generates
   a relative symlink.
   """
+  def symlink_or_copy(path, path) do
+    :ok
+  end
+
   def symlink_or_copy(source, target) do
     if File.exists?(source) do
       # Relative symbolic links on Windows are broken
@@ -644,6 +654,10 @@ defmodule Mix.Utils do
   end
 
   defp read_httpc(path) do
+    Mix.ensure_application!(:public_key)
+    Mix.ensure_application!(:ssl)
+    Mix.ensure_application!(:inets)
+
     {:ok, _} = Application.ensure_all_started(:ssl)
     {:ok, _} = Application.ensure_all_started(:inets)
 
@@ -651,7 +665,7 @@ defmodule Mix.Utils do
     # the effects of using an HTTP proxy to this function
     {:ok, _pid} = :inets.start(:httpc, profile: :mix)
 
-    headers = [{'user-agent', 'Mix/#{System.version()}'}]
+    headers = [{~c"user-agent", ~c"Mix/#{System.version()}"}]
     request = {:binary.bin_to_list(path), headers}
 
     # Use the system certificates if available, otherwise skip peer verification
