@@ -167,6 +167,20 @@ defmodule Mix.Tasks.FormatTest do
     assert output == ""
   end
 
+  test "checks that no error is raised with --check-formatted and --no-exit" do
+    capture_io("foo( )", fn ->
+      Mix.Tasks.Format.run(["--check-formatted", "--no-exit", "-"])
+    end)
+
+    assert_received {:mix_shell, :info, ["The following files are not formatted" <> _]}
+  end
+
+  test "raises an error if --no-exit is passed without --check-formatted" do
+    assert_raise Mix.Error, ~r"--no-exit can only be used together", fn ->
+      Mix.Tasks.Format.run(["--no-exit", "-"])
+    end
+  end
+
   test "uses inputs and configuration from .formatter.exs", context do
     in_tmp(context.test, fn ->
       File.write!(".formatter.exs", """
@@ -258,7 +272,7 @@ defmodule Mix.Tasks.FormatTest do
     def format(contents, opts) do
       assert opts[:from_formatter_exs] == :yes
       assert opts[:sigil] == :W
-      assert opts[:modifiers] == 'abc'
+      assert opts[:modifiers] == ~c"abc"
       assert opts[:line] == 2
       assert opts[:file] =~ ~r/\/a\.ex$/
       contents |> String.split(~r/\s/) |> Enum.join("\n")
@@ -294,6 +308,24 @@ defmodule Mix.Tasks.FormatTest do
                '''abc
              end
              """
+
+      {formatter_function, _options} = Mix.Tasks.Format.formatter_for_file("a.ex")
+
+      assert formatter_function.("""
+             if true do
+               ~W'''
+               foo bar baz
+               '''abc
+             end
+             """) == """
+             if true do
+               ~W'''
+               foo
+               bar
+               baz
+               '''abc
+             end
+             """
     end)
   end
 
@@ -315,6 +347,37 @@ defmodule Mix.Tasks.FormatTest do
     end
   end
 
+  defmodule Elixir.NewlineToDotPlugin do
+    @behaviour Mix.Tasks.Format
+
+    def features(opts) do
+      assert opts[:from_formatter_exs] == :yes
+      [extensions: ~w(.w), sigils: [:W]]
+    end
+
+    def format(contents, opts) do
+      assert opts[:from_formatter_exs] == :yes
+
+      cond do
+        opts[:extension] ->
+          assert opts[:extension] == ".w"
+          assert opts[:file] =~ ~r/\/a\.w$/
+          assert [W: sigil_fun] = opts[:sigils]
+          assert is_function(sigil_fun, 2)
+
+        opts[:sigil] ->
+          assert opts[:sigil] == :W
+          assert opts[:inputs] == ["a.ex"]
+          assert opts[:modifiers] == ~c"abc"
+
+        true ->
+          flunk("Plugin not loading in correctly.")
+      end
+
+      contents |> String.replace("\n", ".")
+    end
+  end
+
   test "uses extension plugins from .formatter.exs", context do
     in_tmp(context.test, fn ->
       File.write!(".formatter.exs", """
@@ -339,6 +402,100 @@ defmodule Mix.Tasks.FormatTest do
     end)
   end
 
+  test "uses multiple plugins from .formatter.exs targeting the same file extension", context do
+    in_tmp(context.test, fn ->
+      File.write!(".formatter.exs", """
+      [
+        inputs: ["a.w"],
+        plugins: [ExtensionWPlugin, NewlineToDotPlugin],
+        from_formatter_exs: :yes
+      ]
+      """)
+
+      File.write!("a.w", """
+      foo bar baz
+      """)
+
+      Mix.Tasks.Format.run([])
+
+      assert File.read!("a.w") == "foo.bar.baz."
+    end)
+  end
+
+  test "uses multiple plugins from .formatter.exs with the same file extension in declared order",
+       context do
+    in_tmp(context.test, fn ->
+      File.write!(".formatter.exs", """
+      [
+        inputs: ["a.w"],
+        plugins: [NewlineToDotPlugin, ExtensionWPlugin],
+        from_formatter_exs: :yes
+      ]
+      """)
+
+      File.write!("a.w", """
+      foo bar baz
+      """)
+
+      Mix.Tasks.Format.run([])
+
+      assert File.read!("a.w") == "foo\nbar\nbaz."
+    end)
+  end
+
+  test "uses multiple plugins from .formatter.exs targeting the same sigil", context do
+    in_tmp(context.test, fn ->
+      File.write!(".formatter.exs", """
+      [
+        inputs: ["a.ex"],
+        plugins: [NewlineToDotPlugin, SigilWPlugin],
+        from_formatter_exs: :yes
+      ]
+      """)
+
+      File.write!("a.ex", """
+      def sigil_test(assigns) do
+        ~W"foo bar baz\n"abc
+      end
+      """)
+
+      Mix.Tasks.Format.run([])
+
+      assert File.read!("a.ex") == """
+             def sigil_test(assigns) do
+               ~W"foo\nbar\nbaz."abc
+             end
+             """
+    end)
+  end
+
+  test "uses multiple plugins from .formatter.exs with the same sigil in declared order",
+       context do
+    in_tmp(context.test, fn ->
+      File.write!(".formatter.exs", """
+      [
+        inputs: ["a.ex"],
+        plugins: [SigilWPlugin, NewlineToDotPlugin],
+        from_formatter_exs: :yes
+      ]
+      """)
+
+      File.write!("a.ex", """
+      def sigil_test(assigns) do
+        ~W"foo bar baz"abc
+      end
+      """)
+
+      Mix.Tasks.Format.run([])
+
+      assert File.read!("a.ex") == """
+             def sigil_test(assigns) do
+               ~W"foo.bar.baz"abc
+             end
+             """
+    end)
+  end
+
   test "uses extension plugins with --stdin-filename", context do
     in_tmp(context.test, fn ->
       File.write!(".formatter.exs", """
@@ -384,6 +541,35 @@ defmodule Mix.Tasks.FormatTest do
     end)
   end
 
+  test "uses inputs and configuration from :root path", context do
+    in_tmp(context.test, fn ->
+      File.write!(".formatter.exs", """
+      [
+        locals_without_parens: [foo: 1]
+      ]
+      """)
+
+      File.mkdir_p!("lib")
+
+      File.write!("lib/a.ex", """
+      foo bar baz
+      """)
+
+      root = File.cwd!()
+
+      {formatter_function, _options} =
+        File.cd!("lib", fn ->
+          Mix.Tasks.Format.formatter_for_file("lib/a.ex", root: root)
+        end)
+
+      assert formatter_function.("""
+             foo bar baz
+             """) == """
+             foo bar(baz)
+             """
+    end)
+  end
+
   test "reads exported configuration from subdirectories", context do
     in_tmp(context.test, fn ->
       File.write!(".formatter.exs", """
@@ -396,10 +582,16 @@ defmodule Mix.Tasks.FormatTest do
       [inputs: "a.ex", locals_without_parens: [my_fun: 2]]
       """)
 
+      # Should hit the formatter
       {formatter, formatter_opts} = Mix.Tasks.Format.formatter_for_file("lib/extra/a.ex")
       assert Keyword.get(formatter_opts, :locals_without_parens) == [my_fun: 2]
       assert formatter.("my_fun 1, 2") == "my_fun 1, 2\n"
 
+      # Another directory should not hit the formatter
+      {formatter, formatter_opts} = Mix.Tasks.Format.formatter_for_file("test/a.ex")
+      assert Keyword.get(formatter_opts, :locals_without_parens) == []
+      assert formatter.("my_fun 1, 2") == "my_fun(1, 2)\n"
+
       File.write!("lib/a.ex", """
       my_fun :foo, :bar
       other_fun :baz
@@ -576,7 +768,7 @@ defmodule Mix.Tasks.FormatTest do
       []
       """)
 
-      message = "Expected :inputs or :subdirectories key in lib/.formatter.exs"
+      message = "Expected :inputs or :subdirectories key in #{Path.expand("lib/.formatter.exs")}"
       assert_raise Mix.Error, message, fn -> Mix.Tasks.Format.run([]) end
     end)
   end
@@ -590,8 +782,7 @@ defmodule Mix.Tasks.FormatTest do
       """)
 
       message =
-        "Unavailable dependency :my_dep given to :import_deps in the formatter configuration. " <>
-          "The dependency cannot be found in the file system, please run \"mix deps.get\" and try again"
+        ~r"Unknown dependency :my_dep given to :import_deps in the formatter configuration"
 
       assert_raise Mix.Error, message, fn -> Mix.Tasks.Format.run([]) end
 
@@ -600,8 +791,7 @@ defmodule Mix.Tasks.FormatTest do
       """)
 
       message =
-        "Unknown dependency :nonexistent_dep given to :import_deps in the formatter configuration. " <>
-          "The dependency is not listed in your mix.exs for environment :dev"
+        ~r"Unknown dependency :nonexistent_dep given to :import_deps in the formatter configuration"
 
       assert_raise Mix.Error, message, fn -> Mix.Tasks.Format.run([]) end
     end)
@@ -633,16 +823,17 @@ defmodule Mix.Tasks.FormatTest do
       Mix.Tasks.Format.run([])
 
       message1 =
-        "Both .formatter.exs and lib/.formatter.exs specify the file lib/a.ex in their " <>
-          ":inputs option. To resolve the conflict, the configuration in .formatter.exs " <>
-          "will be ignored. Please change the list of :inputs in one of the formatter files " <>
-          "so only one of them matches lib/a.ex"
+        "Both .formatter.exs and #{Path.expand("lib/.formatter.exs")} specify the file " <>
+          "#{Path.expand("lib/a.ex")} in their :inputs option. To resolve the conflict, " <>
+          "the configuration in .formatter.exs will be ignored. Please change the list of " <>
+          ":inputs in one of the formatter files so only one of them matches #{Path.expand("lib/a.ex")}"
 
       message2 =
-        "Both lib/.formatter.exs and foo/.formatter.exs specify the file lib/a.ex in their " <>
-          ":inputs option. To resolve the conflict, the configuration in lib/.formatter.exs " <>
+        "Both #{Path.expand("lib/.formatter.exs")} and #{Path.expand("foo/.formatter.exs")} " <>
+          "specify the file #{Path.expand("lib/a.ex")} in their :inputs option. To resolve " <>
+          "the conflict, the configuration in #{Path.expand("lib/.formatter.exs")} " <>
           "will be ignored. Please change the list of :inputs in one of the formatter files " <>
-          "so only one of them matches lib/a.ex"
+          "so only one of them matches #{Path.expand("lib/a.ex")}"
 
       assert_received {:mix_shell, :error, [^message1]}
       assert_received {:mix_shell, :error, [^message2]}
@@ -686,4 +877,382 @@ defmodule Mix.Tasks.FormatTest do
       assert_received {:mix_shell, :error, ["mix format failed for stdin"]}
     end)
   end
+
+  describe "text_diff_format/3" do
+    defp text_diff_format(old, new, opts \\ []) do
+      Mix.Tasks.Format.text_diff_format(old, new, opts) |> IO.iodata_to_binary()
+    end
+
+    test "with unchanged texts" do
+      assert text_diff_format("abc", "abc") == ""
+    end
+
+    test "with one deleted line" do
+      old = "del"
+      new = ""
+
+      assert output = text_diff_format(old, new)
+
+      if IO.ANSI.enabled?() do
+        assert output == "1  \e[31m -\e[0m|\e[31mdel\e[0m\n  1\e[32m +\e[0m|\n"
+      end
+
+      assert text_diff_format(old, new, color: false) == """
+             1   -|del
+               1 +|
+             """
+    end
+
+    test "with one changed line" do
+      old = "one three two"
+      new = "one two three"
+
+      assert output = text_diff_format(old, new)
+
+      if IO.ANSI.enabled?() do
+        assert output ==
+                 """
+                 1  \e[31m -\e[0m|one three\e[31m\e[0m\e[41m \e[0m\e[31mtwo\e[0m
+                   1\e[32m +\e[0m|one t\e[32mwo\e[0m\e[42m \e[0m\e[32mt\e[0mhree
+                 """
+      end
+
+      assert text_diff_format(old, new, color: false) == """
+             1   -|one three two
+               1 +|one two three
+             """
+    end
+
+    test "with one deleted line in the middle" do
+      old = """
+      aaa
+      bbb
+      ccc
+      ddd
+      eee
+      fff
+      ggg
+      """
+
+      new = """
+      aaa
+      bbb
+      ccc
+      eee
+      fff
+      ggg
+      """
+
+      exp = """
+           |
+      2 2  |bbb
+      3 3  |ccc
+      4   -|ddd
+      5 4  |eee
+      6 5  |fff
+           |
+      """
+
+      assert text_diff_format(old, new)
+      assert text_diff_format(old, new, color: false) == exp
+    end
+
+    test "with multiple deleted lines" do
+      old = """
+      aaa
+      bbb
+      ccc
+      ddd
+      eee
+      fff
+      ggg\
+      """
+
+      new = """
+      aaa
+      ggg\
+      """
+
+      exp = """
+      1 1  |aaa
+      2   -|bbb
+      3   -|ccc
+      4   -|ddd
+      5   -|eee
+      6   -|fff
+      7 2  |ggg
+      """
+
+      assert text_diff_format(old, new)
+      assert text_diff_format(old, new, color: false) == exp
+    end
+
+    test "with one added line in the middle" do
+      old = """
+      aaa
+      bbb
+      ccc
+      eee
+      fff
+      ggg
+      """
+
+      new = """
+      aaa
+      bbb
+      ccc
+      ddd
+      eee
+      fff
+      ggg
+      """
+
+      exp = """
+           |
+      2 2  |bbb
+      3 3  |ccc
+        4 +|ddd
+      4 5  |eee
+      5 6  |fff
+           |
+      """
+
+      assert text_diff_format(old, new)
+      assert text_diff_format(old, new, color: false) == exp
+    end
+
+    test "with changed first line" do
+      old = """
+      aaa
+      bbb
+      ccc
+      ddd
+      """
+
+      new = """
+      axa
+      bbb
+      ccc
+      ddd
+      """
+
+      exp = """
+      1   -|aaa
+        1 +|axa
+      2 2  |bbb
+      3 3  |ccc
+           |
+      """
+
+      assert text_diff_format(old, new)
+      assert text_diff_format(old, new, color: false) == exp
+    end
+
+    test "with changed last line" do
+      old = """
+      aaa
+      bbb
+      ccc
+      ddd
+      """
+
+      new = """
+      aaa
+      bbb
+      ccc
+      dxd
+      """
+
+      exp = """
+           |
+      2 2  |bbb
+      3 3  |ccc
+      4   -|ddd
+        4 +|dxd
+      """
+
+      assert text_diff_format(old, new)
+      assert text_diff_format(old, new, color: false) == exp
+    end
+
+    test "with changed first and last line" do
+      old = """
+      aaa
+      bbb
+      ccc
+      ddd
+      eee
+      """
+
+      new = """
+      axa
+      bbb
+      ccc
+      ddd
+      exe
+      """
+
+      exp = """
+      1   -|aaa
+        1 +|axa
+      2 2  |bbb
+           |
+      4 4  |ddd
+      5   -|eee
+        5 +|exe
+      """
+
+      assert text_diff_format(old, new)
+      assert text_diff_format(old, new, color: false, before: 1, after: 1) == exp
+    end
+
+    test "with changed second and second last line" do
+      old = """
+      aaa
+      bbb
+      ccc
+      ddd
+      eee
+      fff
+      ggg
+      hhh
+      iii\
+      """
+
+      new = """
+      aaa
+      bXb
+      ccc
+      ddd
+      eee
+      fff
+      ggg
+      hXh
+      iii\
+      """
+
+      exp = """
+      1 1  |aaa
+      2   -|bbb
+        2 +|bXb
+      3 3  |ccc
+           |
+      7 7  |ggg
+      8   -|hhh
+        8 +|hXh
+      9 9  |iii
+      """
+
+      assert text_diff_format(old, new)
+      assert text_diff_format(old, new, color: false, before: 1, after: 1) == exp
+    end
+
+    test "colorized added tab" do
+      assert output = text_diff_format("ab", "a\tb")
+
+      if IO.ANSI.enabled?() do
+        assert output =~ "\e[42m\t"
+      end
+    end
+
+    test "colorized deleted tab" do
+      assert output = text_diff_format("a\tb", "ab")
+
+      if IO.ANSI.enabled?() do
+        assert output =~ "\e[41m\t"
+      end
+    end
+
+    test "shows added CR" do
+      old = """
+      aaa
+      bbb
+      """
+
+      new = """
+      aaa\r
+      bbb
+      """
+
+      exp = """
+      1   -|aaa
+        1 +|aaa↵
+      2 2  |bbb
+           |
+      """
+
+      assert text_diff_format(old, new)
+      assert text_diff_format(old, new, color: false, before: 1, after: 1) == exp
+    end
+
+    test "shows multiple added CRs" do
+      old = """
+      aaa
+      bbb
+      """
+
+      new = """
+      aaa\r
+      bbb\r
+      ccc\r
+      """
+
+      exp = """
+      1   -|aaa
+      2   -|bbb
+        1 +|aaa↵
+        2 +|bbb↵
+        3 +|ccc\r
+      """
+
+      assert text_diff_format(old, new)
+      assert text_diff_format(old, new, color: false, before: 1, after: 1) == exp
+    end
+
+    test "shows deleted CR" do
+      old = """
+      aaa\r
+      bbb
+      """
+
+      new = """
+      aaa
+      bbb
+      """
+
+      exp = """
+      1   -|aaa↵
+        1 +|aaa
+      2 2  |bbb
+           |
+      """
+
+      assert text_diff_format(old, new)
+      assert text_diff_format(old, new, color: false, before: 1, after: 1) == exp
+    end
+
+    test "shows multiple deleted CRs" do
+      old = """
+      aaa\r
+      bbb\r
+      """
+
+      new = """
+      aaa
+      bbb
+      ccc
+      """
+
+      exp = """
+      1   -|aaa↵
+      2   -|bbb↵
+        1 +|aaa
+        2 +|bbb
+        3 +|ccc
+      """
+
+      assert text_diff_format(old, new)
+      assert text_diff_format(old, new, color: false) == exp
+    end
+  end
 end
